# -*- coding: utf-8 -*-
import abc
import typing

if typing.TYPE_CHECKING:
    from kiara import Kiara


class KiaraRenderer(abc.ABC):
    def __init__(
        self,
        config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        if kiara is None:
            from kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: Kiara = kiara
        if config is None:
            config = {}
        self._config: typing.Mapping[str, typing.Any] = config

    def _augment_inputs(self, **inputs: typing.Any) -> typing.Mapping[str, typing.Any]:
        return inputs

    def _post_process(
        self, rendered: typing.Any, inputs: typing.Mapping[str, typing.Any]
    ) -> typing.Any:
        return rendered

    def render(self, **inputs: typing.Mapping[str, typing.Any]):

        augmented_inputs = self._augment_inputs(**inputs)
        rendered = self._render_template(inputs=augmented_inputs)
        post_processed = self._post_process(rendered=rendered, inputs=augmented_inputs)
        return post_processed

    @abc.abstractmethod
    def _render_template(self, inputs: typing.Mapping[str, typing.Any]):
        pass



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara import Kiara, KiaraModule
from kiara.data import ValueSet
from kiara.data.values import Value, ValueSchema
from kiara.defaults import DEFAULT_PRETTY_PRINT_CONFIG
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message


class PrettyPrintModuleConfig(ModuleTypeConfigSchema):

    value_type: str = Field(description="The type of the value to save.")
    target_type: str = Field(
        description="The target to print the value to.", default="string"
    )


class PrettyPrintValueModule(KiaraModule):

    _config_cls = PrettyPrintModuleConfig

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: Kiara
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        for value_type_name, value_type_cls in kiara.type_mgmt.value_types.items():

            for attr in dir(value_type_cls):
                if not attr.startswith("pretty_print_as_"):
                    continue

                target_type = attr[16:]
                if target_type not in kiara.type_mgmt.value_type_names:
                    log_message(
                        f"Pretty print target type '{target_type}' for source type '{value_type_name}' not valid, ignoring."
                    )
                    continue

                op_config = {
                    "module_type": cls._module_type_id,  # type: ignore
                    "module_config": {
                        "value_type": value_type_name,
                        "target_type": target_type,
                    },
                    "doc": f"Pretty print a value of type '{value_type_name}' as '{target_type}'.",
                }
                all_metadata_profiles[
                    f"{value_type_name}.pretty_print_as.{target_type}"
                ] = op_config

        return all_metadata_profiles

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        inputs: typing.Mapping[str, typing.Any] = {
            "value_item": {
                "type": self.get_config_value("value_type"),
                "doc": f"A value of type '{self.get_config_value('value_type')}'.",
            },
            "print_config": {
                "type": "dict",
                "doc": "Optional print configuration.",
                "optional": True,
            },
        }

        return inputs

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        outputs: typing.Mapping[str, typing.Any] = {
            "printed": {
                "type": self.get_config_value("target_type"),
                "doc": f"The printed value as '{self.get_config_value('target_type')}'.",
            }
        }
        return outputs

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value_obj: Value = inputs.get_value_obj("value_item")
        print_config = dict(DEFAULT_PRETTY_PRINT_CONFIG)
        config: typing.Mapping = inputs.get_value_data("print_config")
        if config:
            print_config.update(config)

        value_type: str = self.get_config_value("value_type")
        target_type: str = self.get_config_value("target_type")

        func_name = f"pretty_print_as_{target_type}"

        if not hasattr(value_obj.type_obj, func_name):
            raise Exception(
                f"Type '{value_type}' can't be pretty printed as '{target_type}'. This is most likely a bug."
            )

        if not value_obj.is_set:
            printed = "-- not set --"
        else:
            func = getattr(value_obj.type_obj, func_name)
            # TODO: check signature
            printed = func(value=value_obj, print_config=print_config)

        outputs.set_value("printed", printed)


class PrettyPrintOperationType(OperationType):
    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, PrettyPrintValueModule)

    def get_pretty_print_operation(
        self, value_type: str, target_type: str
    ) -> Operation:

        result = []
        for op_config in self.operation_configs.values():
            if op_config.module_config["value_type"] != value_type:
                continue
            if op_config.module_config["target_type"] != target_type:
                continue
            result.append(op_config)

        if not result:
            raise Exception(
                f"No pretty print operation for value type '{value_type}' and output '{target_type}' registered."
            )
        elif len(result) != 1:
            raise Exception(
                f"Multiple pretty print operations for value type '{value_type}' and output '{target_type}' registered."
            )

        return result[0]

    def pretty_print(
        self,
        value: Value,
        target_type: str,
        print_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ) -> typing.Any:

        ops_config = self.get_pretty_print_operation(
            value_type=value.type_name, target_type=target_type
        )
        result = ops_config.module.run(value_item=value, print_config=print_config)
        printed = result.get_value_data("printed")

        return printed



# -*- coding: utf-8 -*-

"""This is the base module that contains everything data type-related in *kiara*.

I'm still not 100% sure how to best implement the *kiara* type system, there are several ways it could be done, for example
based on Python type-hints, using JSON-schema, Avro (which is my 2nd favourite option), as well as by implementing a
custom type-class hierarchy. Which is what I have choosen to try first. For now, it looks like it'll work out,
but there is a chance requirements I haven't forseen will crop up that could make this become ugly.

Anyway, the way it works (for now) is that *kiara* comes with a set of often used types (the standard set of: scalars,
list, dict, table & array, etc.) which each come with 2 functions that can serialize and deserialize values of that
type in a persistant fashion -- which could be storing as a file on disk, or as a cell/row in a database. Those functions
will most likley be *kiara* modules themselves, with even more restricted input/output type options.

In addition, packages that contain modules can implement their own, custom types, if suitable ones are not available in
core-*kiara*. Those can either be 'serialized/deserialized' into *kiara*-native types (which in turn will serialize them
using their own serializing functions), or will have to implement custom serializing functionality (which will probably
be discouraged, since this might not be trivial and there are quite a few things to consider).

"""

import deepdiff
import typing
from pydantic import BaseModel, Extra, Field, PrivateAttr, ValidationError
from pydantic.schema import (
    get_flat_models_from_model,
    get_model_name_map,
    model_process_schema,
)
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table

from kiara.defaults import DEFAULT_NO_DESC_VALUE
from kiara.exceptions import KiaraValueException, ValueTypeConfigException
from kiara.metadata import MetadataModel, ValueTypeAndDescription
from kiara.metadata.core_models import PythonClassMetadata
from kiara.metadata.type_models import ValueTypeMetadata


def get_type_name(obj: typing.Any):
    """Utility function to get a pretty string from the class of an object."""

    if obj.__class__.__module__ == "builtins":
        return obj.__class__.__name__
    else:
        return f"{obj.__class__.__module__}.{obj.__class__.__name__}"


class ValueTypeConfigSchema(BaseModel):
    """Base class that describes the configuration a [``ValueType``][kiara.data.types.ValueType] class accepts.

    This is stored in the ``_config_cls`` class attribute in each ``ValueType`` class. By default,
    a ``ValueType`` is not configurable, unless the ``_config_cls`` class attribute points to a sub-class of this class.
    """

    @classmethod
    def requires_config(cls) -> bool:
        """Return whether this class can be used as-is, or requires configuration before an instance can be created."""

        for field_name, field in cls.__fields__.items():
            if field.required and field.default is None:
                return True
        return False

    _config_hash: str = PrivateAttr(default=None)

    class Config:
        extra = Extra.forbid
        allow_mutation = False

    def get(self, key: str) -> typing.Any:
        """Get the value for the specified configuation key."""

        if key not in self.__fields__:
            raise Exception(
                f"No config value '{key}' in module config class '{self.__class__.__name__}'."
            )

        return getattr(self, key)

    @property
    def config_hash(self):

        if self._config_hash is None:
            _d = self.dict()
            hashes = deepdiff.DeepHash(_d)
            self._config_hash = hashes[_d]
        return self._config_hash

    def __eq__(self, other):

        if self.__class__ != other.__class__:
            return False

        return self.dict() == other.dict()

    def __hash__(self):

        return hash(self.config_hash)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        my_table = Table(box=box.MINIMAL, show_header=False)
        my_table.add_column("Field name", style="i")
        my_table.add_column("Value")
        for field in self.__fields__:
            my_table.add_row(field, getattr(self, field))

        yield my_table


TYPE_CONFIG = typing.TypeVar("TYPE_CONFIG", bound=ValueTypeConfigSchema)


class ValueTypeConfigMetadata(MetadataModel):
    @classmethod
    def from_config_class(
        cls,
        config_cls: typing.Type[ValueTypeConfigSchema],
    ):

        flat_models = get_flat_models_from_model(config_cls)
        model_name_map = get_model_name_map(flat_models)
        m_schema, _, _ = model_process_schema(config_cls, model_name_map=model_name_map)
        fields = m_schema["properties"]

        config_values = {}
        for field_name, details in fields.items():
            type_str = "-- n/a --"
            if "type" in details.keys():
                type_str = details["type"]

            desc = details.get("description", DEFAULT_NO_DESC_VALUE)

            default = config_cls.__fields__[field_name].default

            if default is None:
                if callable(config_cls.__fields__[field_name].default_factory):
                    default = config_cls.__fields__[field_name].default_factory()  # type: ignore

            req = config_cls.__fields__[field_name].required

            config_values[field_name] = ValueTypeAndDescription(
                description=desc, type=type_str, value_default=default, required=req
            )

        python_cls = PythonClassMetadata.from_class(config_cls)
        return ValueTypeConfigMetadata(
            python_class=python_cls, config_values=config_values
        )

    python_class: PythonClassMetadata = Field(
        description="The Python class for this configuration."
    )
    config_values: typing.Dict[str, ValueTypeAndDescription] = Field(
        description="The available configuration values."
    )


class ValueType(typing.Generic[TYPE_CONFIG]):
    """Base class that all *kiara* types must inherit from.

    *kiara* types have 3 main responsibilities:

     - serialize into / deserialize from persistent state
     - data validation
     -  metadata extraction

     Serializing being the arguably most important of those, because without most of the data management features of
     *kiara* would be impossible. Validation should not require any explanation. Metadata extraction is important, because
     that metadata will be available to other components of *kiara* (or frontends for it), without them having to request
     the actual data. That will hopefully make *kiara* very efficient in terms of memory management, as well as data
     transfer and I/O. Ideally, the actual data (bytes) will only be requested at the last possible moment. For example when a
     module needs the input data to do processing on it -- and even then it might be that it only requests a part of the
     data, say a single column of a table. Or when a frontend needs to display/visualize the data.
    """

    _config_class: typing.Type[TYPE_CONFIG] = ValueTypeConfigSchema  # type: ignore

    @classmethod
    def get_type_metadata(cls) -> ValueTypeMetadata:
        return ValueTypeMetadata.from_value_type_class(cls)

    # @classmethod
    # def doc(cls) -> str:
    #
    #     return extract_doc_from_cls(cls)
    #
    # @classmethod
    # def desc(cls) -> str:
    #     return extract_doc_from_cls(cls, only_first_line=True)

    # @classmethod
    # def conversions(
    #     self,
    # ) -> typing.Optional[typing.Mapping[str, typing.Mapping[str, typing.Any]]]:
    #     """Return a dictionary of configuration for modules that can transform this type.
    #
    #     The name of the transformation is the key of the result dictionary, the configuration is a module configuration
    #     (dictionary wth 'module_type' and optional 'module_config', 'input_name' and 'output_name' keys).
    #     """
    #
    #     return {"string": {"module_type": "string.pretty_print", "input_name": "item"}}

    @classmethod
    def check_data(cls, data: typing.Any) -> typing.Optional["ValueType"]:
        """Check whether the provided input matches this value type.

        If it does, return a ValueType object (with the appropriate type configuration).
        """
        return None

    @classmethod
    def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
        return None

    @classmethod
    def get_supported_hash_types(cls) -> typing.Iterable[str]:
        return []

    @classmethod
    def calculate_value_hash(cls, value: typing.Any, hash_type: str) -> str:
        """Calculate the hash of this value.

        If a hash can't be calculated, or the calculation of a type is not implemented (yet), this will return None.
        """

        raise Exception(f"Value type '{cls._value_type_name}' does not support hash calculation.")  # type: ignore

    def __init__(self, **type_config: typing.Any):

        try:
            self._type_config = self.__class__._config_class(**type_config)
        except ValidationError as ve:
            raise ValueTypeConfigException(
                f"Error creating object for value_type: {ve}",
                self.__class__,
                type_config,
                ve,
            )
        # self._type_config: typing.Mapping[str, typing.Any] = self
        # self._transformations: typing.Optional[
        #     typing.Mapping[str, typing.Mapping[str, typing.Any]]
        # ] = None

    @property
    def type_config(self) -> TYPE_CONFIG:
        return self._type_config

    def import_value(self, value: typing.Any) -> typing.Any:

        assert value is not None

        try:
            parsed = self.parse_value(value)
            if parsed is None:
                parsed = value
            self.validate(parsed)
        except Exception as e:
            raise KiaraValueException(
                value_type=self.__class__, value_data=value, exception=e
            )

        return parsed

    def parse_value(self, value: typing.Any) -> typing.Any:
        """Parse a value into a supported python type.

        This exists to make it easier to do trivial conversions (e.g. from a date string to a datetime object).
        If you choose to overwrite this method, make 100% sure that you don't change the meaning of the value, and try to
        avoid adding or removing information from the data (e.g. by changing the resolution of a date).

        Arguments:
            v: the value

        Returns:
            'None', if no parsing was done and the original value should be used, otherwise return the parsed Python object
        """
        return None

    def validate(cls, value: typing.Any) -> None:
        pass

    def get_type_hint(self, context: str = "python") -> typing.Optional[typing.Type]:
        """Return a type hint for this value type object.

        This can be used by kiara interfaces to document/validate user input. For now, only 'python' type hints are
        expected to be implemented, but in theory this could also return type hints for other contexts.
        """
        return None

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        raise NotImplementedError()


# class ValueTypeInfo(object):
#     def __init__(self, value_type_cls: typing.Type[ValueType]):
#
#         self._value_type_cls: typing.Type[ValueType] = value_type_cls
#
#     @property
#     def doc(self) -> str:
#         return self._value_type_cls.doc()
#
#     @property
#     def desc(self) -> str:
#         return self._value_type_cls.desc()


class ValueTypesInfo(object):
    def __init__(
        self,
        value_type_classes: typing.Mapping[str, typing.Type[ValueType]],
        details: bool = False,
    ):

        self._value_type_classes: typing.Mapping[
            str, typing.Type[ValueType]
        ] = value_type_classes
        self._details: bool = details

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        table = Table(show_header=True, box=box.SIMPLE, show_lines=False)
        table.add_column("Type name", style="i")
        table.add_column("Description")

        for type_name in sorted(self._value_type_classes.keys()):
            t_md = self._value_type_classes[type_name].get_type_metadata()
            if self._details:
                md = Markdown(t_md.documentation.full_doc)
            else:
                md = Markdown(t_md.documentation.description)
            table.add_row(type_name, md)

        panel = Panel(table, title="Available value types", title_align="left")
        yield panel



# -*- coding: utf-8 -*-
import fnmatch
import re
import typing
from jinja2 import Environment
from pydantic import BaseModel, Field

from kiara.data import Value


class FieldMatcher(BaseModel):
    @classmethod
    def create_default_matchers(cls):

        return [FieldMatcher()]

    match_expr: str = "*"
    matcher_type: str = "glob"
    alias_template: str = "{{ base_id }}__{{ field_name }}"
    check_next_on_match: bool = False


class ValueStoreConfig(BaseModel):
    """A configuration that describes which step outputs of a pipeline to save, and how."""

    @classmethod
    def save_fields(
        self,
        base_id: str,
        value_set: typing.Mapping[str, Value],
        matchers: typing.Optional[typing.List[FieldMatcher]] = None,
    ) -> typing.Dict[Value, typing.List[str]]:

        if matchers is None:
            matchers = FieldMatcher.create_default_matchers()

        for matcher in matchers:
            if matcher.matcher_type not in ["glob", "regex"]:
                raise NotImplementedError(
                    "Only 'glob' and 'regex' field matcher type implemented yet."
                )

            if matcher.matcher_type == "glob":
                matcher.matcher_type = "regex"
                matcher.match_expr = fnmatch.translate(matcher.match_expr)

        env = Environment()

        to_save: typing.Dict[str, typing.Set[str]] = {}
        for field_name in value_set.keys():

            for matcher in matchers:
                if not re.match(matcher.match_expr, field_name):
                    continue

                template = env.from_string(matcher.alias_template)
                rendered = template.render(field_name=field_name, base_id=base_id)

                to_save.setdefault(field_name, set()).add(rendered)
                if not matcher.check_next_on_match:
                    break

        result: typing.Dict[Value, typing.List[str]] = {}
        for field_name, aliases in to_save.items():

            value = value_set[field_name]
            _v = value.save(aliases=aliases)
            result.setdefault(_v, []).extend(aliases)

        return result

    inputs: typing.List[FieldMatcher] = Field(
        description="Whether and how to save inputs.", default_factory=list
    )
    outputs: typing.List[FieldMatcher] = Field(
        description="Whether and how to save inputs.",
        default_factory=FieldMatcher.create_default_matchers,
    )
    steps: typing.List["ValueStoreConfig"] = Field(
        description="Whether and how to save step inputs and outputs.",
        default_factory=list,
    )


ValueStoreConfig.update_forward_refs()



# -*- coding: utf-8 -*-
import os
import typing
from pydantic import BaseSettings, Extra, Field, validator

from kiara.defaults import KIARA_DATA_STORE_DIR, kiara_app_dirs
from kiara.module_mgmt import ModuleManager
from kiara.module_mgmt.pipelines import PipelineModuleManagerConfig
from kiara.module_mgmt.python_classes import PythonModuleManagerConfig
from kiara.processing.parallel import ThreadPoolProcessorConfig
from kiara.processing.synchronous import SynchronousProcessorConfig
from kiara.utils import get_data_from_file

try:
    pass
except Exception:
    pass


def yaml_config_settings_source(settings: BaseSettings) -> typing.Dict[str, typing.Any]:
    """
    A simple settings source that loads variables from a JSON file
    at the project's root.

    Here we happen to choose to use the `env_file_encoding` from Config
    when reading `config.json`
    """

    config_file = os.path.join(kiara_app_dirs.user_config_dir, "config.yaml")
    if os.path.exists(config_file):
        data = get_data_from_file(config_file)
        return data
    else:
        return {}


class KiaraConfig(BaseSettings):
    class Config:
        extra = Extra.forbid
        env_file_encoding = "utf-8"
        env_prefix = "kiara_"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                yaml_config_settings_source,
                file_secret_settings,
            )

    module_managers: typing.Optional[
        typing.List[
            typing.Union[PythonModuleManagerConfig, PipelineModuleManagerConfig]
        ]
    ] = Field(
        description="The module managers to use in this kiara instance.", default=None
    )
    default_processor: typing.Optional[
        typing.Union[SynchronousProcessorConfig, ThreadPoolProcessorConfig]
    ] = Field(
        description="The configuration for the default processor to use.",
        default_factory=SynchronousProcessorConfig,
    )
    data_store: str = Field(
        description="The path to the local kiara data store.",
        default=KIARA_DATA_STORE_DIR,
    )
    extra_pipeline_folders: typing.List[str] = Field(
        description="Paths to local folders that contain kiara pipelines.",
        default_factory=list,
    )
    ignore_errors: bool = Field(
        description="If set, kiara will try to ignore most errors (that can be ignored).",
        default=False,
    )

    @validator("module_managers", pre=True)
    def _validate_managers(cls, v):

        if v is None:
            return []

        if isinstance(v, typing.Mapping):
            v = [v]

        assert isinstance(v, typing.Iterable)

        result = []
        for item in v:
            if isinstance(item, ModuleManager):
                result.append(item)
            else:
                assert isinstance(item, typing.Mapping)
                mm_type = item.get("module_manager_type", None)
                if not mm_type:
                    raise ValueError(
                        f"No module manager type provided in config: {item}"
                    )

                if mm_type == "python":
                    item_config = PythonModuleManagerConfig(**item)
                elif mm_type == "pipeline":
                    item_config = PipelineModuleManagerConfig(**item)
                else:
                    raise ValueError(f"Invalid module manager type: {mm_type}")
                result.append(item_config)

        return result

    @validator("default_processor", pre=True)
    def _validate_default_processor(cls, v):

        if not v:
            return SynchronousProcessorConfig()

        if isinstance(v, (SynchronousProcessorConfig, ThreadPoolProcessorConfig)):
            return v

        if v == "synchronous":
            return SynchronousProcessorConfig()

        if v == "multi-threaded":
            return ThreadPoolProcessorConfig()

        if not isinstance(v, typing.Mapping):
            raise ValueError(
                f"Invalid type '{type(v)}' for default_processor config: {v}"
            )
        processor_type = v.get("module_processor_type", None)
        if not processor_type:
            raise ValueError("No 'module_processor_type' provided: {config}")
        if processor_type == "synchronous":
            config = SynchronousProcessorConfig(**v)
        elif processor_type == "multi-threaded":
            config = ThreadPoolProcessorConfig()

        return config



# -*- coding: utf-8 -*-

"""A module that contains value-related classes for *Kiara*.

A value in Kiara-speak is a pointer to actual data (aka 'bytes'). It contains metadata about that data (like whether it's
valid/set, what type/schema it has, when it was last modified, ...), but it does not contain the data itself. The reason for
that is that such data can be fairly large, and in a lot of cases it is not necessary for the code involved to have
access to it, access to the metadata is enough.

Each Value has a unique id, which can be used to retrieve the data (whole, or parts of it) from a [DataRegistry][kiara.data.registry.DataRegistry].
In addition, that id can be used to subscribe to change events for a value (published whenever the data that is associated with a value was changed).
"""

import json
import logging
import networkx as nx
import typing
import uuid
from datetime import datetime
from pydantic import BaseModel, Extra, Field, PrivateAttr
from rich import box
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.jupyter import JupyterMixin
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.tree import Tree

from kiara.data.types import ValueType
from kiara.defaults import COLOR_LIST, SpecialValue
from kiara.info import KiaraInfoModel
from kiara.metadata.core_models import DocumentationMetadataModel
from kiara.metadata.data import DeserializeConfig
from kiara.module_config import ModuleConfig
from kiara.utils import StringYAML, is_debug, log_message

if typing.TYPE_CHECKING:
    from kiara.data.registry import (
        BaseDataRegistry,
        DataRegistry,
        ValueSlotUpdateHandler,
    )
    from kiara.kiara import Kiara
    from kiara.module import KiaraModule

log = logging.getLogger("kiara")
yaml = StringYAML()


class ValueSchema(BaseModel):
    """The schema of a value.

    The schema contains the [ValueType][kiara.data.values.ValueType] of a value, as well as an optional default that
    will be used if no user input was given (yet) for a value.

    For more complex container types like array, tables, unions etc, types can also be configured with values from the ``type_config`` field.
    """

    class Config:
        use_enum_values = True
        # extra = Extra.forbid

    type: str = Field(description="The type of the value.")
    type_config: typing.Dict[str, typing.Any] = Field(
        description="Configuration for the type, in case it's complex.",
        default_factory=dict,
    )
    default: typing.Any = Field(
        description="A default value.", default=SpecialValue.NOT_SET
    )

    optional: bool = Field(
        description="Whether this value is required (True), or whether 'None' value is allowed (False).",
        default=False,
    )
    is_constant: bool = Field(
        description="Whether the value is a constant.", default=False
    )
    # required: typing.Any = Field(
    #     description="Whether this value is required to be set.", default=True
    # )

    doc: str = Field(
        default="-- n/a --",
        description="A description for the value of this input field.",
    )

    def is_required(self):

        if self.optional:
            return False
        else:
            if self.default in [None, SpecialValue.NOT_SET, SpecialValue.NO_VALUE]:
                return True
            else:
                return False

    def validate_types(self, kiara: "Kiara"):

        if self.type not in kiara.value_type_names:
            raise ValueError(
                f"Invalid value type '{self.type}', available types: {kiara.value_type_names}"
            )

    @property
    def desc(self):
        """The first line of the 'doc' value."""

        return self.doc.split("\n")[0]

    def __eq__(self, other):

        if not isinstance(other, ValueSchema):
            return False

        return (self.type, self.default) == (other.type, other.default)

    def __hash__(self):

        return hash((self.type, self.default))


class ValueLineage(ModuleConfig):
    """Model containing the lineage of a value."""

    @classmethod
    def from_module_and_inputs(
        cls,
        module: "KiaraModule",
        output_name: str,
        inputs: typing.Mapping[str, typing.Union["Value", "ValueInfo"]],
    ):

        module_type = module._module_type_id  # type: ignore
        module_config = module.config.dict()
        doc = module.get_type_metadata().documentation

        _inputs = {}
        for field_name, value in inputs.items():
            if isinstance(value, Value):
                _inputs[field_name] = value.get_info()
            else:
                _inputs[field_name] = value

        return ValueLineage.construct(
            module_type=module_type,
            module_config=module_config,
            doc=doc,
            output_name=output_name,
            inputs=_inputs,
        )

    @classmethod
    def create(
        cls,
        module_type: str,
        module_config: typing.Mapping[str, typing.Any],
        module_doc: DocumentationMetadataModel,
        inputs: typing.Mapping[str, typing.Union["Value", "ValueInfo"]],
        output_name: typing.Optional[str] = None,
    ):

        _inputs = {}
        for field_name, value in inputs.items():
            if isinstance(value, Value):
                _inputs[field_name] = value.get_info()
            else:
                _inputs[field_name] = value

        return ValueLineage.construct(
            module_type=module_type,
            module_config=dict(module_config),
            doc=module_doc,
            output_name=output_name,
            inputs=_inputs,
        )

    output_name: typing.Optional[str] = Field(
        description="The result field name for the value this refers to."
    )
    inputs: typing.Dict[str, "ValueInfo"] = Field(
        description="The inputs that were used to create the value this refers to."
    )
    value_index: typing.Optional[typing.Dict[str, "ValueInfo"]] = Field(
        description="Index of all values that are associated with this value lineage.",
        default=None,
    )

    def to_minimal_dict(
        self,
        include_metadata: bool = False,
        include_module_doc: bool = False,
        include_module_config: bool = True,
    ) -> typing.Dict[str, typing.Any]:

        full_dict = self.dict(exclude_none=True)
        minimal_dict = filter_metadata_schema(
            full_dict,
            include_metadata=include_metadata,
            include_module_doc=include_module_doc,
            include_module_config=include_module_config,
        )
        return minimal_dict

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        all_ids = sorted(find_all_ids_in_lineage(self))
        id_color_map = {}
        for idx, v_id in enumerate(all_ids):
            id_color_map[v_id] = COLOR_LIST[idx % len(COLOR_LIST)]

        show_ids = config.get("include_ids", False)

        tree = fill_lineage_tree(self, include_ids=show_ids)
        return tree

    def create_graph(self) -> nx.DiGraph:

        return create_lineage_graph(self)


def filter_metadata_schema(
    data: typing.Mapping[str, typing.Any],
    include_metadata: bool = False,
    include_module_doc: bool = False,
    include_module_config: bool = True,
) -> typing.Dict[str, typing.Any]:

    result = {}
    for k, v in data.items():

        if (
            isinstance(v, typing.Mapping)
            and "metadata_item" in v.keys()
            and not include_metadata
        ):
            result[k] = v["metadata_item"]
        elif isinstance(v, typing.Mapping):
            if k == "doc" and not include_module_doc:
                continue
            elif k == "module_config" and not include_module_config:
                continue
            else:
                result[k] = filter_metadata_schema(
                    v,
                    include_metadata=include_metadata,
                    include_module_doc=include_module_doc,
                )
        else:
            result[k] = v

    return result


def find_all_ids_in_lineage(lineage: ValueLineage, ids: typing.Set[str] = None):

    if ids is None:
        ids = set()

    if not lineage:
        return ids

    for input_name, value_info in lineage.inputs.items():
        ids.add(value_info.value_id)
        if value_info.lineage:
            find_all_ids_in_lineage(value_info.lineage, ids=ids)

    return ids


def fill_lineage_tree(
    lineage: ValueLineage,
    node: typing.Optional[Tree] = None,
    include_ids: bool = False,
    level=0,
):

    color = COLOR_LIST[level % len(COLOR_LIST)]
    if node is None:
        main = Tree(f"[b {color}]{lineage.module_type}[/b {color}]")
    else:
        main = node.add(f"[b {color}]{lineage.module_type}[/b {color}]")

    for input_name in sorted(lineage.inputs.keys()):

        value_info = lineage.inputs[input_name]

        value_type = value_info.value_schema.type
        if include_ids:
            v_id_str = f" = {value_info.value_id}"
        else:
            v_id_str = ""
        input_node = main.add(
            f"input: [i {color}]{input_name} ({value_type})[/i {color}]{v_id_str}"
        )
        if value_info.lineage:
            fill_lineage_tree(
                value_info.lineage, input_node, level=level + 1, include_ids=include_ids
            )

    return main


def create_lineage_graph(
    lineage: ValueLineage, graph: nx.Graph = None, node=None, level=0
):

    if graph is None:
        assert node is None
        graph = nx.DiGraph()
        # node = graph.add_node(lineage.output_name)
        # module_type_node = graph.add_edge(node, lineage.module_type)
        # for field_name, value in lineage.inputs.items():
        #     graph.add_edge(module_type_node, value.value_id)

    return graph


class ValueHash(BaseModel):

    hash: str = Field(description="The value hash.")
    hash_type: str = Field(description="The value hash method.")


NO_ID_YET_MARKER = "__no_id_yet__"


class Value(BaseModel, JupyterMixin):
    """The underlying base class for all values.

    The important subclasses here are the ones inheriting from 'PipelineValue', as those are registered in the data
    registry.
    """

    class Config:
        extra = Extra.forbid
        use_enum_values = True

    def __init__(self, registry: "BaseDataRegistry", value_schema: ValueSchema, type_obj: ValueType, is_set: bool, is_none: bool, hashes: typing.Optional[typing.Iterable[ValueHash]] = None, metadata: typing.Optional[typing.Mapping[str, typing.Dict[str, typing.Any]]] = None, register_token: typing.Optional[uuid.UUID] = None):  # type: ignore

        if not register_token:
            raise Exception("No register token provided.")

        if not registry._check_register_token(register_token):
            raise Exception(
                f"Value registration with token '{register_token}' not allowed."
            )

        if value_schema is None:
            raise NotImplementedError()

        assert registry
        self._registry = registry
        self._kiara = self._registry._kiara

        kwargs: typing.Dict[str, typing.Any] = {}
        kwargs["id"] = NO_ID_YET_MARKER

        kwargs["value_schema"] = value_schema

        # if value_lineage is None:
        #     value_lineage = ValueLineage()
        #
        # kwargs["value_lineage"] = value_lineage

        # kwargs["is_streaming"] = False  # not used yet
        kwargs["creation_date"] = datetime.now()
        kwargs["is_set"] = is_set
        kwargs["is_none"] = is_none

        if hashes:
            kwargs["hashes"] = list(hashes)

        if metadata:
            kwargs["metadata"] = dict(metadata)
        else:
            kwargs["metadata"] = {}

        super().__init__(**kwargs)
        self._type_obj = type_obj

    _kiara: "Kiara" = PrivateAttr()
    _registry: "BaseDataRegistry" = PrivateAttr()
    _type_obj: ValueType = PrivateAttr()
    _value_info: "ValueInfo" = PrivateAttr(default=None)

    id: str = Field(description="A unique id for this value.")
    value_schema: ValueSchema = Field(description="The schema of this value.")
    creation_date: typing.Optional[datetime] = Field(
        description="The time this value was created value happened."
    )
    # is_streaming: bool = Field(
    #     default=False,
    #     description="Whether the value is currently streamed into this object.",
    # )
    is_set: bool = Field(
        description="Whether the value was set (in some way: user input, default, constant...).",
        default=False,
    )
    is_none: bool = Field(description="Whether the value is 'None'.", default=True)
    hashes: typing.List[ValueHash] = Field(
        description="Available hashes relating to the actual value data. This attribute is not populated by default, use the 'get_hashes()' method to request one or several hash items, afterwards those hashes will be reflected in this attribute.",
        default_factory=list,
    )

    metadata: typing.Dict[str, typing.Dict[str, typing.Any]] = Field(
        description="Available metadata relating to the actual value data (size, no. of rows, etc. -- depending on data type). This attribute is not populated by default, use the 'get_metadata()' method to request one or several metadata items, afterwards those metadata items will be reflected in this attribute.",
        default_factory=dict,
    )

    @property
    def type_name(self) -> str:
        return self.value_schema.type

    @property
    def type_obj(self):
        """Return the object that contains all the type information for this value."""
        return self._type_obj

    def get_hash(self, hash_type: str) -> ValueHash:
        """Calculate the hash of a specified type for this value. Hashes are cached."""

        hashes = self.get_hashes(hash_type)
        return list(hashes)[0]

    def get_hashes(self, *hash_types: str) -> typing.Iterable[ValueHash]:

        all_hash_types = self.type_obj.get_supported_hash_types()
        if not hash_types:
            try:
                hash_types = all_hash_types
            except Exception as e:
                log_message(str(e))

            if not hash_types:
                return []

        result = []
        missing = list(hash_types)
        for hash_obj in self.hashes:
            if hash_obj.hash_type in hash_types:
                result.append(hash_obj)
                missing.remove(hash_obj.hash_type)

        for hash_type in missing:
            if hash_type not in all_hash_types:
                raise Exception(
                    f"Hash type '{hash_type}' not supported for '{self.type_name}'"
                )

            hash_str = self.type_obj.calculate_value_hash(
                value=self.get_value_data(), hash_type=hash_type
            )
            hash_obj = ValueHash(hash_type=hash_type, hash=hash_str)
            self.hashes.append(hash_obj)
            result.append(hash_obj)

        return result

    def item_is_valid(self) -> bool:
        """Check whether the current value is valid"""

        if self.value_schema.optional:
            return True
        else:
            return self.is_set and not self.is_none

    def item_status(self) -> str:
        """Print a human readable short description of this values status."""

        if self.value_schema.optional:
            if self.is_set:
                if self.is_none:
                    return "no value (not required)"
                else:
                    if self.value_schema.default and self.type_name in [
                        "string",
                        "integer",
                        "boolean",
                    ]:
                        if self.get_value_data() == self.value_schema.default:
                            return "set (default)"
                    return "set"
            else:
                return "not set (not required)"
        else:
            if self.is_set:
                if self.is_none:
                    return "no value"
                else:
                    if self.value_schema.default and self.type_name in [
                        "string",
                        "integer",
                        "boolean",
                    ]:
                        if self.get_value_data() == self.value_schema.default:
                            return "set (default)"
                    return "set"
            else:
                return "not set"

    def get_value_data(self) -> typing.Any:

        return self._registry.get_value_data(self)

    def get_lineage(self) -> typing.Optional[ValueLineage]:

        return self._registry.get_lineage(self)

    def set_value_lineage(self, value_lineage: ValueLineage) -> None:

        if hasattr(self._registry, "set_value_lineage"):
            return self._registry.set_value_lineage(self, value_lineage)  # type: ignore
        else:
            raise Exception("Can't set value lineage: registry is read only")

    def get_info(self) -> "ValueInfo":

        if self._value_info is None:
            self._value_info = ValueInfo.from_value(self)
        return self._value_info

    def create_info(self, include_deserialization_config: bool = False) -> "ValueInfo":

        return ValueInfo.from_value(
            self, include_deserialization_config=include_deserialization_config
        )

    def get_metadata(
        self, *metadata_keys: str, also_return_schema: bool = False
    ) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:
        """Retrieve (if necessary) and return metadata for the specified keys.

        By default, the metadata is returned as a map (in case of a single key: single-item map) with metadata key as
        dict key, and the metadata as value. If 'also_return_schema' was set to `True`, the value will be a two-key
        map with the metadata under the ``metadata_item`` subkey, and the value schema under ``metadata_schema``.

        If no metadata key is specified, all available metadata for this value type will be returned.

        Arguments:
            metadata_keys: the metadata keys to retrieve metadata (empty for all available metadata)
            also_return_schema: whether to also return the schema for each metadata item
        """

        if not metadata_keys:
            _metadata_keys = set(
                self._kiara.metadata_mgmt.get_metadata_keys_for_type(self.type_name)
            )
            for key in self.metadata.keys():
                _metadata_keys.add(key)
        else:
            _metadata_keys = set(metadata_keys)

        result = {}

        missing = set()
        for metadata_key in sorted(_metadata_keys):
            if metadata_key in self.metadata.keys():
                if also_return_schema:
                    result[metadata_key] = self.metadata[metadata_key]
                else:
                    result[metadata_key] = self.metadata[metadata_key]["metadata_item"]
            else:
                missing.add(metadata_key)

        if not missing:
            return result

        _md = self._kiara.metadata_mgmt.get_value_metadata(
            self, *missing, also_return_schema=True
        )
        for k, v in _md.items():
            self.metadata[k] = v
            if also_return_schema:
                result[k] = v
            else:
                result[k] = v["metadata_item"]
        return {k: result[k] for k in sorted(result.keys())}

    def save(
        self,
        aliases: typing.Optional[typing.Iterable[str]] = None,
        register_missing_aliases: bool = True,
    ) -> "Value":
        """Save this value, under the specified alias(es)."""

        # if self.get_value_lineage():
        #     for field_name, value in self.get_value_lineage().inputs.items():
        #         value_obj = self._registry.get_value_obj(value)
        #         try:
        #             value_obj.save()
        #         except Exception as e:
        #             print(e)

        value = self._kiara.data_store.register_data(self)
        if aliases:
            self._kiara.data_store.link_aliases(
                value, *aliases, register_missing_aliases=register_missing_aliases
            )

        return value

    def __eq__(self, other):

        # TODO: compare all attributes if id is equal, just to make sure...

        if not isinstance(other, Value):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):

        return f"Value(id={self.id}, type={self.type_name}, is_set={self.is_set}, is_none={self.is_none}"

    def __str__(self):

        return self.__repr__()

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        # table = self._create_value_table()

        title = "Value"
        yield Panel(self.get_info(), box=box.ROUNDED, title_align="left", title=title)


ValueSchema.update_forward_refs()
Value.update_forward_refs()


class ValueSlot(BaseModel):
    @classmethod
    def from_value(cls, id: str, value: Value) -> "ValueSlot":

        vs = ValueSlot.from_value_schema(
            id=id, value_schema=value.value_schema, kiara=value._kiara
        )
        vs.add_value(value)
        return vs

    @classmethod
    def from_value_schema(
        cls, id: str, value_schema: ValueSchema, kiara: "Kiara"
    ) -> "ValueSlot":

        vs = ValueSlot(id=id, value_schema=value_schema, kiara=kiara)
        return vs

    def __init__(self, **data):  # type: ignore
        _kiara = data.pop("kiara", None)
        if _kiara is None:
            raise Exception("No kiara context provided.")
        _registry = data.pop("registry", None)
        if _registry is None:
            _registry = _kiara.data_registry

        self._kiara = _kiara
        self._registry = _registry
        super().__init__(**data)

    _kiara: "Kiara" = PrivateAttr()
    _registry: "DataRegistry" = PrivateAttr()
    _callbacks: typing.Dict[str, "ValueSlotUpdateHandler"] = PrivateAttr(
        default_factory=dict
    )

    id: str = Field(description="The id for this slot.")
    value_schema: ValueSchema = Field(
        description="The schema for the values of this slot."
    )
    values: typing.Dict[int, Value] = Field(
        description="The values of this slot, with versions as key.",
        default_factory=dict,
    )
    tags: typing.Dict[str, int] = Field(
        description="The tags for this value slot (tag name as key, linked version as value.",
        default_factory=dict,
    )

    @property
    def latest_version_nr(self) -> int:
        if not self.values:
            return 0
        return max(self.values.keys())

    def get_latest_value(self) -> Value:

        lv = self.latest_version_nr
        if lv == 0:
            raise Exception("No value added to value slot yet.")

        return self.values[self.latest_version_nr]

    def register_callbacks(self, *callbacks: "ValueSlotUpdateHandler"):

        for cb in callbacks:
            cb_id: typing.Optional[str] = None
            if cb_id in self._callbacks.keys():
                raise Exception(f"Callback with id '{cb_id}' already registered.")
            if hasattr(cb, "id"):
                if callable(cb.id):  # type: ignore
                    cb_id = cb.id()  # type: ignore
                else:
                    cb_id = cb.id  # type: ignore
            elif hasattr(cb, "get_id"):
                cb_id = cb.get_id()  # type: ignore

            if cb_id is None:
                cb_id = str(uuid.uuid4())

            assert isinstance(cb_id, str)

            self._callbacks[cb_id] = cb

    def add_value(
        self,
        value: Value,
        trigger_callbacks: bool = True,
        tags: typing.Optional[typing.Iterable[str]] = None,
    ) -> int:
        """Add a value to this slot."""

        if self.latest_version_nr != 0 and value.id == self.get_latest_value().id:
            return self.latest_version_nr

        # TODO: check value data equality
        if self.value_schema.is_constant and self.values:
            if is_debug():
                import traceback

                traceback.print_stack()

            raise Exception("Can't add value: value slot marked as 'constant'.")

        version = self.latest_version_nr + 1
        assert version not in self.values.keys()
        self.values[version] = value
        if tags:
            for tag in tags:
                self.tags[tag] = version

        if trigger_callbacks:
            for cb in self._callbacks.values():
                cb.values_updated(self)

        return version

    def is_latest_value(self, value: Value):

        return value.id == self.get_latest_value().id

    def find_linked_aliases(
        self, value_item: typing.Union[Value, str]
    ) -> typing.List["ValueAlias"]:

        if isinstance(value_item, Value):
            value_item = value_item.id

        result = []
        for _version, _value in self.values.items():
            if _value.id == value_item:
                va = ValueAlias(alias=self.id, version=_version)
                result.append(va)
                if _version in self.tags.values():
                    for _tag, _tag_version in self.tags.items():
                        if _tag_version == _version:
                            va = ValueAlias(alias=self.id, tag=_tag)
                            result.append(va)

        return result

    def __eq__(self, other):

        if not isinstance(other, ValueSlot):
            return False

        return self.id == other.id

    def __hash__(self):

        return hash(self.id)


class ValueAlias(BaseModel):
    @classmethod
    def from_string(
        self, value_alias: str, default_repo_name: typing.Optional[str] = None
    ) -> "ValueAlias":

        if not isinstance(value_alias, str):
            raise Exception("Invalid id_or_alias: not a string.")
        if not value_alias:
            raise Exception("Invalid id_or_alias: can't be empty string.")

        _repo_name: typing.Optional[str] = default_repo_name
        _version: typing.Optional[int] = None
        _tag: typing.Optional[str] = None

        if "#" in value_alias:
            _repo_name, _value_alias = value_alias.split("#", maxsplit=1)
        else:
            _value_alias = value_alias

        if "@" in _value_alias:
            _alias, _postfix = _value_alias.split("@", maxsplit=1)

            try:
                _version = int(_postfix)
            except ValueError:
                if not _postfix.isidentifier():
                    raise Exception(
                        f"Invalid format for version/tag element of id_or_alias: {_tag}"
                    )
                _tag = _postfix
        else:
            _alias = _value_alias

        return ValueAlias(
            repo_name=_repo_name, alias=_alias, version=_version, tag=_tag
        )

    @classmethod
    def from_strings(
        cls, *value_aliases: typing.Union[str, "ValueAlias"]
    ) -> typing.List["ValueAlias"]:

        result = []
        for va in value_aliases:
            if isinstance(va, str):
                result.append(ValueAlias.from_string(va))
            elif isinstance(va, ValueAlias):
                result.append(va)
            else:
                raise TypeError(
                    f"Invalid type '{type(va)}' for type alias, expected 'str' or 'ValueAlias'."
                )
        return result

    repo_name: typing.Optional[str] = Field(
        description="The name of the data repo the value lives in.", default=None
    )
    alias: str = Field("The alias name.")
    version: typing.Optional[int] = Field(
        description="The version of this alias.", default=None
    )
    tag: typing.Optional[str] = Field(
        description="The tag for the alias.", default=None
    )

    @property
    def full_alias(self):
        if self.tag is not None:
            return f"{self.alias}@{self.tag}"
        elif self.version is not None:
            return f"{self.alias}@{self.version}"
        else:
            return self.alias


class ValueInfo(KiaraInfoModel):
    @classmethod
    def from_value(cls, value: Value, include_deserialization_config: bool = False):

        if value.id not in value._registry.value_ids:
            raise Exception("Value not registered (yet).")

        # aliases = value._registry.find_aliases_for_value(value)
        hashes = value.get_hashes()
        metadata = value.get_metadata(also_return_schema=True)
        # metadata = value.metadata
        value_lineage = value.get_lineage()

        if include_deserialization_config:
            # serialize_operation: SerializeValueOperationType = (  # type: ignore
            #     value._kiara.operation_mgmt.get_operation("serialize")  # type: ignore
            # )
            raise NotImplementedError()
        return ValueInfo(
            value_id=value.id,
            value_schema=value.value_schema,
            hashes=hashes,
            metadata=metadata,
            lineage=value_lineage,
            is_valid=value.item_is_valid(),
        )

    value_id: str = Field(description="The value id.")
    value_schema: ValueSchema = Field(description="The value schema.")
    # aliases: typing.List[ValueAlias] = Field(
    #     description="All aliases for this value.", default_factory=list
    # )
    # tags: typing.List[str] = Field(
    #     description="All tags for this value.", default_factory=list
    # )
    # created: str = Field(description="The time the data was created.")
    is_valid: bool = Field(
        description="Whether the item is valid (in the context of its schema)."
    )
    hashes: typing.List[ValueHash] = Field(
        description="All available hashes for this value.", default_factory=list
    )
    metadata: typing.Dict[str, typing.Dict[str, typing.Any]] = Field(
        description="The metadata associated with this value."
    )
    lineage: typing.Optional[ValueLineage] = Field(
        description="Information about how the value was created.", default=None
    )
    deserialize_config: typing.Optional[DeserializeConfig] = Field(
        description="The module config (incl. inputs) to deserialize the value.",
        default=None,
    )

    def get_metadata_items(self, *keys: str) -> typing.Dict[str, typing.Any]:

        if not keys:
            _keys: typing.Iterable[str] = self.metadata.keys()
        else:
            _keys = keys

        result = {}
        for k in _keys:

            md = self.metadata.get(k)
            if md is None:
                raise Exception(f"No metadata for key '{k}' available.")

            result[k] = md["metadata_item"]

        return result

    def get_metadata_schemas(self, *keys: str) -> typing.Dict[str, typing.Any]:

        if not keys:
            _keys: typing.Iterable[str] = self.metadata.keys()
        else:
            _keys = keys

        result = {}
        for k in _keys:
            md = self.metadata.get(k)
            if md is None:
                raise Exception(f"No metadata for key '{k}' available.")
            result[k] = md["metadata_schema"]

        return result

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        padding = config.get("padding", (0, 1))
        skip_metadata = config.get("skip_metadata", False)
        skip_value_lineage = config.get("skip_lineage", True)
        include_ids = config.get("include_ids", False)

        table = Table(box=box.SIMPLE, show_header=False, padding=padding)
        table.add_column("property", style="i")
        table.add_column("value")

        table.add_row("id", self.value_id)  # type: ignore
        table.add_row("type", self.value_schema.type)
        if self.value_schema.type_config:
            json_data = json.dumps(self.value_schema.type_config)
            tc_content = Syntax(json_data, "json")
            table.add_row("type config", tc_content)
        table.add_row("desc", self.value_schema.doc)
        table.add_row("is set", "yes" if self.is_valid else "no")
        # table.add_row("is constant", "yes" if self.is_constant else "no")

        # if isinstance(self.value_hash, int):
        #     vh = str(self.value_hash)
        # else:
        #     vh = self.value_hash.value
        # table.add_row("hash", vh)

        if self.hashes:
            hashes_dict = {hs.hash_type: hs.hash for hs in self.hashes}
            yaml_string = yaml.dump(hashes_dict)
            hases_str = Syntax(yaml_string, "yaml", background_color="default")
            table.add_row("", "")
            table.add_row("hashes", hases_str)

        if not skip_metadata:
            if self.metadata:
                yaml_string = yaml.dump(data=self.get_metadata_items())
                # json_string = json.dumps(self.get_metadata_items(), indent=2)
                metadata = Syntax(yaml_string, "yaml", background_color="default")
                table.add_row("metadata", metadata)
            else:
                table.add_row("metadata", "-- no metadata --")

        if not skip_value_lineage and self.lineage:
            if self.metadata:
                table.add_row("", "")
            # json_string = self.lineage.json(indent=2)
            # seed_content = Syntax(json_string, "json")
            table.add_row(
                "lineage", self.lineage.create_renderable(include_ids=include_ids)
            )

        return table


ValueLineage.update_forward_refs()



# -*- coding: utf-8 -*-
"""Pipeline-related subcommands for the cli."""
import click
import os.path
import sys
import typing
from rich.panel import Panel

from kiara import Kiara
from kiara.info.modules import ModuleTypesGroupInfo
from kiara.info.pipelines import PipelineModuleInfo
from kiara.utils import log_message
from kiara.utils.output import rich_print


@click.group()
@click.pass_context
def pipeline(ctx):
    """Pipeline-related sub-commands."""


@pipeline.command(name="list")
@click.option(
    "--full-doc",
    "-d",
    is_flag=True,
    help="Display the full documentation for every module type.",
)
@click.argument("filter", nargs=-1, required=False)
@click.pass_context
def list_pipelines(
    ctx,
    full_doc: bool,
    filter: typing.Iterable[str],
):
    """List available module types."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    if filter:
        module_types = []

        for m in kiara_obj.available_pipeline_module_types:
            match = True

            for f in filter:

                if f.lower() not in m.lower():
                    match = False
                    break
                else:
                    m_cls = kiara_obj.get_module_class(m)
                    doc = m_cls.get_type_metadata().documentation.full_doc

                    if f.lower() not in doc.lower():
                        match = False
                        break

            if match:
                module_types.append(m)
    else:
        module_types = kiara_obj.available_pipeline_module_types

    renderable = ModuleTypesGroupInfo.create_renderable_from_type_names(
        kiara=kiara_obj,
        type_names=module_types,
        ignore_non_pipeline_modules=True,
        ignore_pipeline_modules=False,
        include_full_doc=full_doc,
    )
    title = "Available pipeline modules"

    p = Panel(renderable, title_align="left", title=title)
    print()
    kiara_obj.explain(p)


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def explain(ctx, pipeline_type: str):
    """Print details about pipeline inputs, outputs, and overall structure."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)
    print()
    kiara_obj.explain(info.structure)


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def explain_steps(ctx, pipeline_type: str):
    """List all steps of a pipeline."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(
        module_type_name=pipeline_type, kiara=kiara_obj
    )
    print()
    st_info = info.structure.steps_info
    rich_print(st_info)


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.pass_context
def execution_graph(ctx, pipeline_type: str):
    """Print the execution graph for a pipeline structure."""

    kiara_obj = ctx.obj["kiara"]

    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)
    info.print_execution_graph()


@pipeline.command()
@click.argument("pipeline-type", nargs=1)
@click.option(
    "--full",
    "-f",
    is_flag=True,
    help="Display full data-flow graph, incl. intermediate input/output connections.",
)
@click.pass_context
def data_flow_graph(ctx, pipeline_type: str, full: bool):
    """Print the data flow graph for a pipeline structure."""

    kiara_obj = ctx.obj["kiara"]
    if os.path.isfile(pipeline_type):
        pipeline_type = kiara_obj.register_pipeline_description(
            pipeline_type, raise_exception=True
        )

    m_cls = kiara_obj.get_module_class(pipeline_type)
    if not m_cls.is_pipeline():
        rich_print()
        rich_print(f"Module '{pipeline_type}' is not a pipeline-type module.")
        sys.exit(1)

    info = PipelineModuleInfo.from_type_name(pipeline_type, kiara=kiara_obj)

    info.print_data_flow_graph(simplified=not full)


try:
    pass

    @pipeline.command()
    @click.argument("pipeline", nargs=1)
    @click.option(
        "--template",
        "-t",
        help="The template to use. Defaults to 'notebook'.",
        default="notebook",
    )
    @click.pass_context
    def render(ctx, pipeline, template):

        kiara_obj: Kiara = ctx.obj["kiara"]

        # pipeline = "/home/markus/projects/dharpa/kiara-playground/examples/streamlit/geolocation_prototype/pipelines/geolocation_1.yml"
        # template = os.path.join(KIARA_RESOURCES_FOLDER, "templates", "python_script.py.j2")
        rendered = kiara_obj.template_mgmt.render(
            "pipeline", module=pipeline, template=template
        )

        print(rendered)


except Exception:
    log_message(
        "'black' or 'jupytext' not installed, not adding 'pipeline render' subcommand."
    )



# -*- coding: utf-8 -*-
from mkdocstrings.handlers.base import BaseCollector, BaseHandler, BaseRenderer


def get_handler():
    return KiaraModuleHandler()


class KiaraModuleCollector(BaseCollector):
    def __init__(self):
        pass


class KiaraModuleRenderer(BaseRenderer):
    def __init__(self):
        pass


class KiaraModuleHandler(BaseHandler):
    def __init__(self):

        collector = KiaraModuleCollector()
        renderer = KiaraModuleRenderer()
        super().__init__(collector=collector, renderer=renderer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for kiara.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
import pytest

import os
import tempfile
import typing
import uuid

from kiara.config import KiaraConfig
from kiara.kiara import Kiara
from kiara.pipeline.config import PipelineConfig

from .utils import INVALID_PIPELINES_FOLDER, MODULE_CONFIGS_FOLDER, PIPELINES_FOLDER

TEMP_DIR = os.path.join(tempfile.gettempdir(), "kiara_tests")
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


@pytest.fixture
def pipeline_paths():

    result = {}
    for root, dirnames, filenames in os.walk(PIPELINES_FOLDER, topdown=True):

        for f in filenames:
            full = os.path.join(root, f)
            if os.path.isfile(full) and f.endswith(".json"):
                result[os.path.splitext(f)[0]] = full

    return result


@pytest.fixture
def invalid_pipeline_paths():

    result = {}
    for root, dirnames, filenames in os.walk(INVALID_PIPELINES_FOLDER, topdown=True):

        for f in filenames:
            full = os.path.join(root, f)
            if os.path.isfile(full) and f.endswith(".json"):
                result[os.path.splitext(f)[0]] = full

    return result


@pytest.fixture
def module_config_paths():

    result = {}
    for root, dirnames, filenames in os.walk(MODULE_CONFIGS_FOLDER, topdown=True):

        for f in filenames:
            full = os.path.join(root, f)
            if os.path.isfile(full) and f.endswith(".json"):
                result[os.path.splitext(f)[0]] = full

    return result


@pytest.fixture
def pipeline_configs(pipeline_paths) -> typing.Mapping[str, PipelineConfig]:

    return {
        name: PipelineConfig.parse_file(path) for name, path in pipeline_paths.items()
    }


@pytest.fixture
def kiara() -> Kiara:

    kiara = Kiara()
    return kiara


@pytest.fixture(scope="module")
def presseeded_data_store():

    session_id = str(uuid.uuid4())

    instance_data_store = os.path.join(TEMP_DIR, f"instance_{session_id}")
    conf = KiaraConfig(data_store=instance_data_store)

    pipeline_file = os.path.join(PIPELINES_FOLDER, "table_import.json")

    kiara = Kiara(config=conf)
    kiara.register_pipeline_description(pipeline_file, module_type_name="preseed")
    results = kiara.run(
        "preseed",
        inputs={
            "import_file__source": os.path.join(
                ROOT_DIR, "examples", "data", "journals", "JournalNodes1902.csv"
            )
        },
    )

    table_value = results.get_value_obj("create_table_from_files__value_item")
    table_value.save(aliases=["journal_nodes"])

    return kiara



# -*- coding: utf-8 -*-
import threading


class ThreadSaveCounter(object):
    """A thread-safe counter, can be used in kiara modules to update completion percentage."""

    def __init__(self):

        self._current = 0
        self._lock = threading.Lock()

    @property
    def current(self):
        return self._current

    def current_percent(self, total: int) -> int:

        return int((self.current / total) * 100)

    def increment(self):

        with self._lock:
            self._current += 1
            return self._current

    def decrement(self):

        with self._lock:
            self._current -= 1
            return self._current



# -*- coding: utf-8 -*-
"""Module related subcommands for the cli."""
import click
import os.path
import sys
import typing
from rich.panel import Panel

from kiara import Kiara
from kiara.info.modules import ModuleTypesGroupInfo
from kiara.interfaces.cli.utils import _create_module_instance
from kiara.metadata.module_models import KiaraModuleTypeMetadata
from kiara.utils import dict_from_cli_args, is_debug, log_message
from kiara.utils.output import rich_print


@click.group()
@click.pass_context
def module(ctx):
    """Module-related sub-commands.."""


@module.command(name="list")
@click.option(
    "--only-pipeline-modules", "-p", is_flag=True, help="Only list pipeline modules."
)
@click.option(
    "--only-core-modules",
    "-c",
    is_flag=True,
    help="Only list core (aka 'Python') modules.",
)
@click.option(
    "--full-doc",
    "-d",
    is_flag=True,
    help="Display the full documentation for every module type.",
)
@click.argument("filter", nargs=-1, required=False)
@click.pass_context
def list_modules(
    ctx,
    only_pipeline_modules: bool,
    only_core_modules: bool,
    full_doc: bool,
    filter: typing.Iterable[str],
):
    """List available module types."""

    if only_pipeline_modules and only_core_modules:
        rich_print()
        rich_print(
            "Please provide either '--only-core-modules' or '--only-pipeline-modules', not both."
        )
        sys.exit(1)

    kiara_obj: Kiara = ctx.obj["kiara"]

    if filter:
        module_types = []

        for m in kiara_obj.available_module_types:
            match = True

            for f in filter:

                if f.lower() not in m.lower():
                    match = False
                    break
                else:
                    m_cls = kiara_obj.get_module_class(m)
                    doc = m_cls.get_type_metadata().documentation.full_doc

                    if f.lower() not in doc.lower():
                        match = False
                        break

            if match:
                module_types.append(m)
    else:
        module_types = kiara_obj.available_module_types

    renderable = ModuleTypesGroupInfo.create_renderable_from_type_names(
        kiara=kiara_obj,
        type_names=module_types,
        ignore_non_pipeline_modules=only_pipeline_modules,
        ignore_pipeline_modules=only_core_modules,
        include_full_doc=full_doc,
    )
    if only_pipeline_modules:
        title = "Available pipeline modules"
    elif only_core_modules:
        title = "Available core modules"
    else:
        title = "Available modules"

    p = Panel(renderable, title_align="left", title=title)
    print()
    kiara_obj.explain(p)


@module.command(name="explain")
@click.argument("module_type", nargs=1, required=True)
@click.pass_context
def explain_module_type(ctx, module_type: str):
    """Print details of a module type.

    This is different to the 'explain-instance' command, because module types need to be
    instantiated with configuration, before we can query all their properties (like
    input/output types).
    """

    kiara_obj: Kiara = ctx.obj["kiara"]

    if os.path.isfile(module_type):
        _module_type: str = kiara_obj.register_pipeline_description(  # type: ignore
            module_type, raise_exception=True
        )  # type: ignore
    else:
        _module_type = module_type

    m_cls = kiara_obj.get_module_class(_module_type)
    info = KiaraModuleTypeMetadata.from_module_class(m_cls)

    rich_print()
    rich_print(info.create_panel(title=f"Module type: [b i]{module_type}[/b i]"))


@module.command("explain-instance")
@click.argument("module_type", nargs=1)
@click.argument(
    "module_config",
    nargs=-1,
)
@click.pass_context
def explain_module(ctx, module_type: str, module_config: typing.Iterable[typing.Any]):
    """Describe a module instance.

    This command shows information and metadata about an instantiated *kiara* module.
    """

    if module_config:
        module_config = dict_from_cli_args(*module_config)
    else:
        module_config = {}

    module_obj = _create_module_instance(
        ctx, module_type=module_type, module_config=module_config
    )
    rich_print()
    rich_print(module_obj)


try:

    from kiara_streamlit.defaults import (  # type: ignore
        MODULE_DEV_STREAMLIT_FILE,
        MODULE_INFO_UI_STREAMLIT_FILE,
    )
    from kiara_streamlit.utils import run_streamlit
    from streamlit.cli import configurator_options

    @module.command("dev")
    @configurator_options
    @click.argument("module_name", required=False, nargs=1)
    @click.argument("args", nargs=-1)
    @click.pass_context
    def dev_ui(ctx, module_name, args=None, **kwargs):
        """Auto-render web-ui to help with module development.

        If no module name is provided, a selection box will displayed in the published app.

        This subcommand uses [streamlit](https://streamlit.io) to auto-render a UI for a (single) module, incl. input fields,
        input previews, output previews, and debug messages. Its main purpose is to aid module development, but it can be
        used as a module execution UI in a pinch.
        """

        kiara_obj: Kiara = ctx.obj["kiara"]

        run_streamlit(
            kiara=kiara_obj,
            streamlit_app_path=MODULE_DEV_STREAMLIT_FILE,
            module_name=module_name,
            streamlit_flags=kwargs,
        )

    @module.command("info-ui")
    @configurator_options
    @click.argument("module_name", required=False, nargs=1)
    # @click.argument("args", nargs=-1)
    @click.pass_context
    def info_ui(ctx, module_name, **kwargs):
        """Auto-render web-ui to help with module development.

        If no module name is provided, a selection box will displayed in the published app.

        This subcommand uses [streamlit](https://streamlit.io) to auto-render a UI for a (single) module, incl. input fields,
        input previews, output previews, and debug messages. Its main purpose is to aid module development, but it can be
        used as a module execution UI in a pinch.
        """

        kiara_obj: Kiara = ctx.obj["kiara"]

        run_streamlit(
            kiara=kiara_obj,
            streamlit_app_path=MODULE_INFO_UI_STREAMLIT_FILE,
            module_name=module_name,
            streamlit_flags=kwargs,
        )


except Exception as e:  # noqa
    if is_debug():
        import traceback

        traceback.print_exc()
    log_message(
        "'kiara.streamlit' package not installed, not offering streamlit debug sub-command"
    )

# if is_develop():
#     try:
#         from jupytext import jupytext
#
#         @module.command("render")
#         @click.argument("module_type", nargs=1)
#         @click.argument("inputs", nargs=-1, required=False)
#         @click.option(
#             "--module-config",
#             "-c",
#             required=False,
#             help="(Optional) module configuration.",
#             multiple=True,
#         )
#         @click.pass_context
#         def render(
#             ctx,
#             module_type: str,
#             module_config: typing.Iterable[typing.Any],
#             inputs: typing.Any,
#         ):
#             """Render a workflow into a jupyter notebook."""
#
#             if module_config:
#                 module_config = dict_from_cli_args(*module_config)
#
#             module_obj: PipelineModule = _create_module_instance(  # type: ignore
#                 ctx, module_type=module_type, module_config=module_config
#             )
#             if not module_obj.is_pipeline():
#                 print("Only pipeline modules supported (for now).")
#                 sys.exit(1)
#
#             structure = module_obj.structure
#
#             list_keys = []
#             for name, value_schema in module_obj.input_schemas.items():
#                 if value_schema.type in ["array", "list"]:
#                     list_keys.append(name)
#             workflow_input = dict_from_cli_args(*inputs, list_keys=list_keys)
#
#             renderer = PipelineRenderer(structure=structure)
#             path = os.path.join(
#                 KIARA_RESOURCES_FOLDER, "templates", "notebook.ipynb.j2"
#             )
#             # path = os.path.join(KIARA_RESOURCES_FOLDER, "templates", "python_script.py.j2")
#
#             step_inputs: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
#             for k, v in workflow_input.items():
#                 pi = structure.pipeline_inputs.get(k)
#                 assert pi
#                 if len(pi.connected_inputs) != 1:
#                     raise NotImplementedError()
#
#                 ci = pi.connected_inputs[0]
#                 if isinstance(v, str):
#                     v = f'"{v}"'
#                 step_inputs.setdefault(ci.step_id, {})[ci.value_name] = v
#
#             rendered = renderer.render_from_path(path, inputs=step_inputs)
#             print()
#             # print(rendered)
#             # return
#             # print(rendered)
#             notebook = jupytext.reads(rendered, fmt="py:percent")
#             converted = jupytext.writes(notebook, fmt="notebook")
#             print(converted)
#
#     except ModuleNotFoundError:
#         pass



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara import KiaraModule
from kiara.data import Value, ValueSet
from kiara.data.values import ValueSchema
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara


class CalculateValueHashesConfig(ModuleTypeConfigSchema):

    value_type: str = Field(
        description="The type of the value to calculate the hash for."
    )
    hash_type: str = Field(description="The hash type.")


class CalculateValueHashModule(KiaraModule):
    """Calculate the hash of a value."""

    _module_type_name = "value.hash"
    _config_cls = CalculateValueHashesConfig

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: "Kiara"
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        for value_type_name, value_type in kiara.type_mgmt.value_types.items():

            hash_types = value_type.get_supported_hash_types()

            for ht in hash_types:
                op_config = {
                    "module_type": cls._module_type_id,  # type: ignore
                    "module_config": {"value_type": value_type_name, "hash_type": ht},
                    "doc": f"Calculate '{ht}' hash for value type '{value_type_name}'.",
                }
                all_metadata_profiles[
                    f"{value_type_name}.calculate_hash.{ht}"
                ] = op_config

        return all_metadata_profiles

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        return {
            "value_item": {
                "type": self.get_config_value("value_type"),
                "doc": f"A value of type '{self.get_config_value('value_type')}'.",
            }
        }

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        return {"hash": {"type": "string", "doc": "The hash string."}}

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value: Value = inputs.get_value_obj("value_item")

        value_hash = value.get_hash(hash_type=self.get_config_value("hash_type"))
        outputs.set_value("hash", value_hash)


class CalculateHashOperationType(OperationType):
    """Calculate a hash for a dataset."""

    def is_matching_operation(self, op_config: Operation) -> bool:

        return op_config.module_cls == CalculateValueHashModule

    def get_hash_operations_for_type(
        self, value_type: str
    ) -> typing.Dict[str, Operation]:

        result = {}
        for op_config in self.operation_configs.values():
            if op_config.module_config["value_type"] == value_type:
                result[op_config.module_config["hash_type"]] = op_config

        return result



# -*- coding: utf-8 -*-
import logging
import typing
from pydantic import BaseModel, Extra, Field, root_validator, validator
from rich.console import Console, ConsoleOptions, RenderResult
from slugify import slugify

from kiara.module_config import ModuleConfig, ModuleTypeConfigSchema
from kiara.pipeline import StepValueAddress
from kiara.pipeline.structure import PipelineStep, PipelineStructure
from kiara.pipeline.utils import ensure_step_value_addresses

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.pipeline.controller import PipelineController
    from kiara.pipeline.module import PipelineModule

log = logging.getLogger("kiara")


class PipelineStepConfig(ModuleConfig):
    """A class to hold the configuration of one module within a [PipelineModule][kiara.pipeline.module.PipelineModule]."""

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    step_id: str = Field(description="The id of the step.")
    input_links: typing.Dict[str, typing.List[StepValueAddress]] = Field(
        default_factory=dict,
        description="The map with the name of an input link as key, and the connected module output name(s) as value.",
    )

    @root_validator(pre=True)
    def create_step_id(cls, values):

        if "module_type" not in values:
            raise ValueError("No 'module_type' specified.")
        if "step_id" not in values or not values["step_id"]:
            values["step_id"] = slugify(values["module_type"])

        return values

    @validator("step_id")
    def ensure_valid_id(cls, v):

        # TODO: check with regex
        if "." in v or " " in v:
            raise ValueError(
                f"Step id can't contain special characters or whitespaces: {v}"
            )

        return v

    @validator("module_config", pre=True)
    def ensure_dict(cls, v):

        if v is None:
            v = {}
        return v

    @validator("input_links", pre=True)
    def ensure_input_links_valid(cls, v):

        if v is None:
            v = {}

        result = {}
        for input_name, output in v.items():

            input_links = ensure_step_value_addresses(
                default_field_name=input_name, link=output
            )
            result[input_name] = input_links

        return result


# class PipelineStructureConfig(BaseModel):
#
#     parent_id: str = Field(description="The id of the parent of this structure.")
#     steps: typing.List[PipelineStepConfig] = Field(description="The steps and their connections that define this pipeline.")
#     input_aliases: typing.Union[None, str, typing.Dict[str, str]] = Field(description="A map that allows to rename the auto-generated inputs of this pipeline.", default=None)
#     output_aliases: typing.Union[None, str, typing.Dict[str, str]] = Field(description="A map that allows to rename the auto-generated outputs of this pipeline.", default=None)


class PipelineConfig(ModuleTypeConfigSchema):
    """A class to hold the configuration for a [PipelineModule][kiara.pipeline.module.PipelineModule].

    If you want to control the pipeline input and output names, you need to have to provide a map that uses the
    autogenerated field name ([step_id]__[field_name] -- 2 underscores!!) as key, and the desired field name
    as value. The reason that schema for the autogenerated field names exist is that it's hard to ensure
    the uniqueness of each field; some steps can have the same input field names, but will need different input
    values. In some cases, some inputs of different steps need the same input. Those sorts of things.
    So, to make sure that we always use the right values, I chose to implement a conservative default approach,
    accepting that in some cases the user will be prompted for duplicate inputs for the same value.

    To remedy that, the pipeline creator has the option to manually specify a mapping to rename some or all of
    the input/output fields.

    Further, because in a lot of cases there won't be any overlapping fields, the creator can specify ``auto``,
    in which case *Kiara* will automatically create a mapping that tries to map autogenerated field names
    to the shortest possible names for each case.

    Examples:

        Configuration for a pipeline module that functions as a ``nand`` logic gate (in Python):

        ``` python
        and_step = PipelineStepConfig(module_type="and", step_id="and")
        not_step = PipelineStepConfig(module_type="not", step_id="not", input_links={"a": ["and.y"]}
        nand_p_conf = PipelineConfig(doc="Returns 'False' if both inputs are 'True'.",
                            steps=[and_step, not_step],
                            input_aliases={
                                "and__a": "a",
                                "and__b": "b"
                            },
                            output_aliases={
                                "not__y": "y"
                            }}
        ```

        Or, the same thing in json:

        ``` json
        {
          "module_type_name": "nand",
          "doc": "Returns 'False' if both inputs are 'True'.",
          "steps": [
            {
              "module_type": "and",
              "step_id": "and"
            },
            {
              "module_type": "not",
              "step_id": "not",
              "input_links": {
                "a": "and.y"
              }
            }
          ],
          "input_aliases": {
            "and__a": "a",
            "and__b": "b"
          },
          "output_aliases": {
            "not__y": "y"
          }
        }
        ```
    """

    # @classmethod
    # def from_file(cls, path: typing.Union[str, Path]):
    #
    #     content = get_data_from_file(path)
    #     return PipelineConfig(**content)

    @classmethod
    def create_pipeline_config(
        cls,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        kiara: typing.Optional["Kiara"] = None,
    ) -> "PipelineConfig":
        """Create a PipelineModule instance.

        The main 'config' argument here can be either:

          - a string: in which case it needs to be (in that order):
            - a module id
            - an operation id
            - a path to a local file
          - a [ModuleConfig][kiara.module_config.ModuleConfig] object
          - a dict (to create a `ModuleInstsanceConfig` from


        Arguments:
            kiara: the kiara context
            config: the 'main' config object
            module_config: in case the 'main' config object was a module id, this argument is used to instantiate the module
            kiara: the kiara context (will use default context instance if not provided)

        """

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        module_config_obj = ModuleConfig.create_module_config(
            config=config, module_config=module_config, kiara=kiara
        )

        if not module_config_obj.module_type == "pipeline":
            raise Exception(f"Not a valid pipeline configuration: {config}")

        # TODO: this is a bit round-about, to create a module config first, but it probably doesn't matter
        pipeline_config_data = module_config_obj.module_config

        module: PipelineConfig = PipelineConfig(**pipeline_config_data)
        return module

    class Config:
        extra = Extra.allow
        validate_assignment = True

    steps: typing.List[PipelineStepConfig] = Field(
        default_factory=list,
        description="A list of steps/modules of this pipeline, and their connections.",
    )
    input_aliases: typing.Union[str, typing.Dict[str, str]] = Field(
        default_factory=dict,
        description="A map of input aliases, with the calculated (<step_id>__<input_name> -- double underscore!) name as key, and a string (the resulting workflow input alias) as value. Check the documentation for the config class for which marker strings can be used to automatically create this map if possible.",
    )
    output_aliases: typing.Union[str, typing.Dict[str, str]] = Field(
        default_factory=dict,
        description="A map of output aliases, with the calculated (<step_id>__<output_name> -- double underscore!) name as key, and a string (the resulting workflow output alias) as value.  Check the documentation for the config class for which marker strings can be used to automatically create this map if possible.",
    )
    documentation: str = Field(
        default="-- n/a --", description="Documentation about what the pipeline does."
    )

    context: typing.Dict[str, typing.Any] = Field(
        default_factory=dict, description="Metadata for this workflow."
    )

    @validator("steps", pre=True)
    def _validate_steps(cls, v):

        steps = []
        for step in v:
            if isinstance(step, PipelineStepConfig):
                steps.append(step)
            elif isinstance(step, typing.Mapping):
                steps.append(PipelineStepConfig(**step))
            else:
                raise TypeError(step)
        return steps

    def create_pipeline_structure(
        self, kiara: typing.Optional["Kiara"] = None
    ) -> "PipelineStructure":
        from kiara import Kiara, PipelineStructure

        if kiara is None:
            kiara = Kiara.instance()

        ps = PipelineStructure(
            config=self,
            kiara=kiara,
        )
        return ps

    def create_pipeline(
        self,
        controller: typing.Optional["PipelineController"] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        # if parent_id is None:
        #     parent_id = DEFAULT_PIPELINE_PARENT_ID
        structure = self.create_pipeline_structure(kiara=kiara)

        from kiara import Pipeline

        pipeline = Pipeline(
            structure=structure,
            controller=controller,
        )
        return pipeline

    def create_pipeline_module(
        self,
        module_id: typing.Optional[str] = None,
        parent_id: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
    ) -> "PipelineModule":

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        from kiara.pipeline.module import PipelineModule

        module = PipelineModule(
            id=module_id,
            parent_id=parent_id,
            module_config=self,
            kiara=kiara,
        )
        return module

    # def __rich_console__(
    #     self, console: Console, options: ConsoleOptions
    # ) -> RenderResult:
    #
    #     table = Table(show_header=False, box=box.SIMPLE)


class StepDesc(BaseModel):
    """Details of a single [PipelineStep][kiara.pipeline.structure.PipelineStep] (which lives within a [Pipeline][kiara.pipeline.pipeline.Pipeline]"""

    class Config:
        allow_mutation = False
        extra = Extra.forbid

    step: PipelineStep = Field(description="Attributes of the step itself.")
    processing_stage: int = Field(
        description="The processing stage of this step within a Pipeline."
    )
    input_connections: typing.Dict[str, typing.List[str]] = Field(
        description="""A map that explains what elements connect to this steps inputs. A connection could either be a Pipeline input (indicated by the ``__pipeline__`` token), or another steps output.

Example:
``` json
input_connections: {
    "a": ["__pipeline__.a"],
    "b": ["step_one.a"]
}

```
        """
    )
    output_connections: typing.Dict[str, typing.List[str]] = Field(
        description="A map that explains what elemnts connect to this steps outputs. A connection could be either a Pipeline output, or another steps input."
    )
    required: bool = Field(
        description="Whether this step is always required, or potentially could be skipped in case some inputs are not available."
    )

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        yield f"[b]Step: {self.step.step_id}[\b]"



# -*- coding: utf-8 -*-
"""Implementation of interfaces for *Kiara*."""
# -*- coding: utf-8 -*-
import os
from rich.console import Console
from typing import Optional

# Global console used by alternative print
_console: Optional[Console] = None


def get_console() -> Console:
    """Get a global Console instance.

    Returns:
        Console: A console instance.
    """
    global _console
    if _console is None or True:
        console_width = os.environ.get("CONSOLE_WIDTH", None)
        width = None

        if console_width:
            try:
                width = int(console_width)
            except Exception:
                pass

        _console = Console(width=width)

    return _console



# -*- coding: utf-8 -*-
import json
import typing
from pydantic import Field, PrivateAttr
from rich import box
from rich.console import RenderableType, RenderGroup
from rich.syntax import Syntax
from rich.table import Table

from kiara.data import ValueSet
from kiara.data.values import ValueSchema
from kiara.metadata.operation_models import OperationsMetadata
from kiara.module_config import ModuleConfig
from kiara.utils import create_table_from_field_schemas

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.module import KiaraModule


class ClassAttributes(object):
    def __init__(self, attrs: typing.Mapping[str, typing.Type]):
        self._attrs: typing.Iterable[str] = attrs


class Operation(ModuleConfig):
    @classmethod
    def create_operation(
        cls,
        kiara: "Kiara",
        operation_id: str,
        config: typing.Union["ModuleConfig", typing.Mapping, str],
        module_config: typing.Union[None, typing.Mapping[str, typing.Any]] = None,
    ) -> "Operation":

        _config = ModuleConfig.create_module_config(
            config=config, module_config=module_config, kiara=kiara
        )
        _config_dict = _config.dict()
        _config_dict["id"] = operation_id
        op_config = cls(**_config_dict)
        op_config._kiara = kiara
        return op_config

    _kiara: typing.Optional["Kiara"] = PrivateAttr(default=None)
    _module: typing.Optional["KiaraModule"]

    id: str = Field(description="The operation id.")

    @property
    def kiara(self) -> "Kiara":
        if self._kiara is None:
            raise Exception("Kiara context not set for operation.")
        return self._kiara

    @property
    def module(self) -> "KiaraModule":
        if self._module is None:
            self._module = self.create_module(kiara=self.kiara)
        return self._module

    @property
    def input_schemas(self) -> typing.Mapping[str, ValueSchema]:
        return self.module.input_schemas

    @property
    def output_schemas(self) -> typing.Mapping[str, ValueSchema]:
        return self.module.output_schemas

    @property
    def module_cls(self) -> typing.Type["KiaraModule"]:
        return self.kiara.get_module_class(self.module_type)

    def run(self, _attach_lineage: bool = True, **inputs: typing.Any) -> ValueSet:

        return self.module.run(_attach_lineage=_attach_lineage, **inputs)

    def create_renderable(self, **config: typing.Any) -> RenderableType:
        """Create a printable overview of this operations details.

        Available config options:
          - 'include_full_doc' (default: True): whether to include the full documentation, or just a description
          - 'include_src' (default: False): whether to include the module source code
        """

        include_full_doc = config.get("include_full_doc", True)

        table = Table(box=box.SIMPLE, show_header=False, show_lines=True)
        table.add_column("Property", style="i")
        table.add_column("Value")

        if self.doc:
            if include_full_doc:
                table.add_row("Documentation", self.doc.full_doc)
            else:
                table.add_row("Description", self.doc.description)

        module_type_md = self.module.get_type_metadata()

        table.add_row("Module type", self.module_type)
        conf = Syntax(
            json.dumps(self.module_config, indent=2), "json", background_color="default"
        )
        table.add_row("Module config", conf)

        constants = self.module_config.get("constants")
        inputs_table = create_table_from_field_schemas(
            _add_required=True,
            _add_default=True,
            _show_header=True,
            _constants=constants,
            **self.module.input_schemas,
        )
        table.add_row("Inputs", inputs_table)
        outputs_table = create_table_from_field_schemas(
            _add_required=False,
            _add_default=False,
            _show_header=True,
            _constants=None,
            **self.module.output_schemas,
        )
        table.add_row("Outputs", outputs_table)

        m_md_o = module_type_md.origin.create_renderable()
        m_md_c = module_type_md.context.create_renderable()
        m_md = RenderGroup(m_md_o, m_md_c)
        table.add_row("Module metadata", m_md)

        if config.get("include_src", False):
            table.add_row("Source code", module_type_md.process_src)

        return table


class OperationType(object):
    @classmethod
    def get_type_metadata(cls) -> OperationsMetadata:

        return OperationsMetadata.from_operations_class(cls)

    def __init__(self, kiara: typing.Optional["Kiara"] = None):

        if kiara is None:
            from kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: "Kiara" = kiara
        self._operations: typing.Dict[str, Operation] = {}

    def is_matching_operation(self, op_config: Operation) -> bool:
        raise NotImplementedError()

    def add_operation(self, module_type_id, op_config) -> bool:

        if not self.is_matching_operation(op_config):
            return False

        self._operations[module_type_id] = op_config
        return True

    @property
    def operation_configs(self) -> typing.Mapping[str, Operation]:
        return self._operations


class AllOperationType(OperationType):
    def is_matching_operation(self, op_config: Operation) -> bool:
        return True


class OperationMgmt(object):
    def __init__(
        self,
        kiara: "Kiara",
        operation_type_classes: typing.Optional[
            typing.Mapping[str, typing.Type[OperationType]]
        ] = None,
    ):

        self._kiara: "Kiara" = kiara

        self._operation_type_classes: typing.Optional[
            typing.Dict[str, typing.Type["OperationType"]]
        ] = None

        if operation_type_classes is not None:
            self._operation_type_classes = dict(operation_type_classes)

        self._operation_types: typing.Optional[typing.Dict[str, OperationType]] = None

        self._profiles: typing.Optional[typing.Dict[str, Operation]] = None
        self._operations: typing.Optional[typing.Dict[str, typing.List[str]]] = None

    @property
    def operation_type_classes(
        self,
    ) -> typing.Mapping[str, typing.Type["OperationType"]]:

        if self._operation_type_classes is not None:
            return self._operation_type_classes

        from kiara.utils.class_loading import find_all_operation_types

        self._operation_type_classes = find_all_operation_types()
        return self._operation_type_classes

    @property
    def operation_ids(self) -> typing.List[str]:
        return list(self.profiles.keys())

    @property
    def profiles(self) -> typing.Mapping[str, Operation]:

        if self._profiles is not None:
            return self._profiles

        _profiles = {}

        for module_id in self._kiara.available_module_types:

            mod_cls = self._kiara.get_module_class(module_id)
            mod_conf = mod_cls._config_cls
            if not mod_conf.requires_config():
                _profiles[module_id] = Operation.create_operation(
                    operation_id=module_id,
                    config={
                        "module_type": module_id,
                        "doc": mod_cls.get_type_metadata().documentation,
                    },
                    kiara=self._kiara,
                )

            profiles = mod_cls.retrieve_module_profiles(kiara=self._kiara)
            if profiles:
                for profile_name, config in profiles.items():
                    if "." not in profile_name:
                        profile_id = f"{module_id}.{profile_name}"
                    else:
                        profile_id = profile_name

                    if profile_id in _profiles.keys():
                        raise Exception(
                            f"Duplicate operation id '{profile_id}': {_profiles[profile_id].dict()} -- {config}"
                        )

                    if not isinstance(config, Operation):
                        config = Operation.create_operation(
                            operation_id=profile_id, config=config, kiara=self._kiara
                        )
                    _profiles[profile_id] = config

        self._profiles = {k: _profiles[k] for k in sorted(_profiles.keys())}
        _operations: typing.Dict[str, typing.List[str]] = {}

        for profile_name, op_config in self._profiles.items():
            for op_type_name, op_type in self.operation_types.items():
                if op_type.add_operation(profile_name, op_config):
                    _operations.setdefault(op_type_name, []).append(profile_name)

        self._operations = _operations

        return self._profiles

    def get_operation(self, operation_id: str) -> typing.Optional[Operation]:

        return self.profiles.get(operation_id, None)

    @property
    def operation_types(self) -> typing.Mapping[str, OperationType]:

        if self._operation_types is not None:
            return self._operation_types

        # TODO: support op type config
        _operation_types = {}
        for op_name, op_cls in self.operation_type_classes.items():
            _operation_types[op_name] = op_cls(kiara=self._kiara)

        self._operation_types = _operation_types
        # make sure the profiles are loaded
        self.profiles  # noqa
        return self._operation_types

    def get_operations(self, operation_type: str) -> OperationType:

        if operation_type not in self.operation_types.keys():
            raise Exception(f"No operation type '{operation_type}' registered.")
        return self.operation_types[operation_type]

    def get_types_for_id(self, operation_id: str) -> typing.Set[str]:

        result = set()
        for ops_ty, ops in self.operation_types.items():
            if operation_id in ops.operation_configs.keys():
                result.add(ops_ty)
        return result

    def run(self):
        pass



# -*- coding: utf-8 -*-
import typing

from kiara.data import Value
from kiara.data.types import ValueType
from kiara.utils.class_loading import find_all_value_types

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara


class TypeMgmt(object):
    def __init__(self, kiara: "Kiara"):

        self._kiara: Kiara = kiara
        self._value_types: typing.Optional[
            typing.Dict[str, typing.Type[ValueType]]
        ] = None
        self._value_type_transformations: typing.Dict[
            str, typing.Dict[str, typing.Any]
        ] = {}
        self._registered_python_classes: typing.Dict[typing.Type, typing.List[str]] = None  # type: ignore

    def invalidate_types(self):

        self._value_types = None
        self._value_type_transformations.clear()
        self._registered_python_classes = None

    @property
    def value_types(self) -> typing.Mapping[str, typing.Type[ValueType]]:

        if self._value_types is not None:
            return self._value_types

        self._value_types = find_all_value_types()
        return self._value_types

    @property
    def value_type_names(self) -> typing.List[str]:
        return list(self.value_types.keys())

    @property
    def registered_python_classes(
        self,
    ) -> typing.Mapping[typing.Type, typing.Iterable[str]]:

        if self._registered_python_classes is not None:
            return self._registered_python_classes

        registered_types = {}
        for name, v_type in self.value_types.items():
            rel = v_type.candidate_python_types()
            if rel:
                for cls in rel:
                    registered_types.setdefault(cls, []).append(name)

        self._registered_python_classes = registered_types
        return self._registered_python_classes

    def determine_type(self, data: typing.Any) -> typing.Optional[ValueType]:

        if isinstance(data, Value):
            data = data.get_value_data()

        result: typing.List[ValueType] = []

        registered_types = set(self.registered_python_classes.get(data.__class__, []))
        for cls in data.__class__.__bases__:
            reg = self.registered_python_classes.get(cls)
            if reg:
                registered_types.update(reg)

        if registered_types:
            for rt in registered_types:
                _cls: typing.Type[ValueType] = self.get_value_type_cls(rt)
                match = _cls.check_data(data)
                if match:
                    result.append(match)

        # TODO: re-run all checks on all modules, not just the ones that registered interest in the class

        if len(result) == 0:
            return None
        elif len(result) > 1:
            result_str = [x._value_type_name for x in result]  # type: ignore
            raise Exception(
                f"Multiple value types found for value: {', '.join(result_str)}."
            )
        else:
            return result[0]

    def get_value_type_cls(self, type_name: str) -> typing.Type[ValueType]:

        t = self.value_types.get(type_name, None)
        if t is None:
            raise Exception(
                f"No value type '{type_name}', available types: {', '.join(self.value_types.keys())}"
            )
        return t

    # def get_value_type_transformations(
    #     self, value_type_name: str
    # ) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:
    #     """Return available transform pipelines for value types."""
    #
    #     if value_type_name in self._value_type_transformations.keys():
    #         return self._value_type_transformations[value_type_name]
    #
    #     type_cls = self.get_value_type_cls(type_name=value_type_name)
    #     _configs = type_cls.conversions()
    #     if _configs is None:
    #         module_configs = {}
    #     else:
    #         module_configs = dict(_configs)
    #     for base in type_cls.__bases__:
    #         if hasattr(base, "conversions"):
    #             _b_configs = base.conversions()  # type: ignore
    #             if not _b_configs:
    #                 continue
    #             for k, v in _b_configs.items():
    #                 if k not in module_configs.keys():
    #                     module_configs[k] = v
    #
    #     # TODO: check input type compatibility?
    #     result: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
    #     for name, config in module_configs.items():
    #         config = dict(config)
    #         module_type = config.pop("module_type", None)
    #         if not module_type:
    #             raise Exception(
    #                 f"Can't create transformation '{name}' for type '{value_type_name}', no module type specified in config: {config}"
    #             )
    #         module_config = config.pop("module_config", {})
    #         module = self._kiara.create_module(
    #             id=f"_transform_{value_type_name}_{name}",
    #             module_type=module_type,
    #             module_config=module_config,
    #         )
    #
    #         if "input_name" not in config.keys():
    #
    #             if len(module.input_schemas) == 1:
    #                 config["input_name"] = next(iter(module.input_schemas.keys()))
    #             else:
    #                 required_inputs = [
    #                     inp
    #                     for inp, schema in module.input_schemas.items()
    #                     if schema.is_required()
    #                 ]
    #                 if len(required_inputs) == 1:
    #                     config["input_name"] = required_inputs[0]
    #                 else:
    #                     raise Exception(
    #                         f"Can't create transformation '{name}' for type '{value_type_name}': can't determine input name between those options: '{', '.join(required_inputs)}'"
    #                     )
    #
    #         if "output_name" not in config.keys():
    #
    #             if len(module.input_schemas) == 1:
    #                 config["output_name"] = next(iter(module.output_schemas.keys()))
    #             else:
    #                 required_outputs = [
    #                     inp
    #                     for inp, schema in module.output_schemas.items()
    #                     if schema.is_required()
    #                 ]
    #                 if len(required_outputs) == 1:
    #                     config["output_name"] = required_outputs[0]
    #                 else:
    #                     raise Exception(
    #                         f"Can't create transformation '{name}' for type '{value_type_name}': can't determine output name between those options: '{', '.join(required_outputs)}'"
    #                     )
    #
    #         result[name] = {
    #             "module": module,
    #             "module_type": module_type,
    #             "module_config": module_config,
    #             "transformation_config": config,
    #         }
    #
    #     self._value_type_transformations[value_type_name] = result
    #     return self._value_type_transformations[value_type_name]

    def find_value_types_for_package(
        self, package_name: str
    ) -> typing.Dict[str, typing.Type[ValueType]]:

        result = {}
        for value_type_name, value_type in self.value_types.items():

            value_md = value_type.get_type_metadata()
            package = value_md.context.labels.get("package")
            if package == package_name:
                result[value_type_name] = value_type

        return result

    # def get_available_transformations_for_type(
    #     self, value_type_name: str
    # ) -> typing.Iterable[str]:
    #
    #     return self.get_value_type_transformations(value_type_name=value_type_name)

    # def transform_value(
    #     self,
    #     transformation_alias: str,
    #     value: Value,
    #     other_inputs: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    # ) -> Value:
    #
    #     transformations = self.get_value_type_transformations(value.value_schema.type)
    #
    #     if transformation_alias not in transformations.keys():
    #         raise Exception(
    #             f"Can't transform value of type '{value.value_schema.type}', transformation '{transformation_alias}' not available for this type. Available: {', '.join(transformations.keys())}"
    #         )
    #
    #     config = transformations[transformation_alias]
    #
    #     transformation_config = config["transformation_config"]
    #     input_name = transformation_config["input_name"]
    #
    #     module: KiaraModule = config["module"]
    #
    #     constants = module.get_config_value("constants")
    #     inputs = dict(constants)
    #
    #     if other_inputs:
    #
    #         for k, v in other_inputs.items():
    #             if k in constants.keys():
    #                 raise Exception(f"Invalid value '{k}' for 'other_inputs', this is a constant that can't be overwrittern.")
    #             inputs[k] = v
    #
    #     defaults = transformation_config.get("defaults", None)
    #     if defaults:
    #         for k, v in defaults.items():
    #             if k in constants.keys():
    #                 raise Exception(f"Invalid  default value '{k}', this is a constant that can't be overwrittern.")
    #             if k not in inputs.keys():
    #                 inputs[k] = v
    #
    #     if input_name in inputs.keys():
    #         raise Exception(
    #             f"Invalid value for inputs in transform arguments, can't contain the main input key '{input_name}'."
    #         )
    #
    #     inputs[input_name] = value
    #
    #     result = module.run(**inputs)
    #     output_name = transformation_config["output_name"]
    #
    #     result_value = result.get_value_obj(output_name)
    #     return result_value



# -*- coding: utf-8 -*-
import typing

from kiara.data.values import ValueSchema
from kiara.data.values.value_set import ValueSet
from kiara.exceptions import KiaraProcessingException
from kiara.module import KiaraModule
from kiara.pipeline import StepStatus
from kiara.pipeline.config import PipelineConfig
from kiara.pipeline.structure import PipelineStructure
from kiara.utils import StringYAML

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara

yaml = StringYAML()


class PipelineModule(KiaraModule[PipelineConfig]):
    """A [KiaraModule][kiara.module.KiaraModule] that contains a collection of interconnected other modules."""

    _config_cls: typing.Type[PipelineConfig] = PipelineConfig  # type: ignore
    _module_type_id = "pipeline"

    @classmethod
    def is_pipeline(cls) -> bool:
        return True

    def __init__(
        self,
        id: typing.Optional[str],
        parent_id: typing.Optional[str] = None,
        module_config: typing.Union[
            None, PipelineConfig, typing.Mapping[str, typing.Any]
        ] = None,
        # controller: typing.Union[
        #     None, PipelineController, str, typing.Type[PipelineController]
        # ] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        # if controller is not None and not isinstance(controller, PipelineController):
        #     raise NotImplementedError()
        # if controller is None:
        super().__init__(
            id=id,
            parent_id=parent_id,
            module_config=module_config,
            kiara=kiara,
        )
        self._pipeline_structure: PipelineStructure = self._create_structure()
        assert not self._config.constants
        self._config.constants = dict(self._pipeline_structure.constants)

    @property
    def structure(self) -> PipelineStructure:
        """The ``PipelineStructure`` of this module."""

        return self._pipeline_structure

    def _create_structure(self) -> PipelineStructure:

        pipeline_structure = PipelineStructure(config=self.config, kiara=self._kiara)
        return pipeline_structure

    def create_input_schema(self) -> typing.Mapping[str, ValueSchema]:
        return self.structure.pipeline_input_schema

    def create_output_schema(self) -> typing.Mapping[str, ValueSchema]:
        return self.structure.pipeline_output_schema

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        from kiara import Pipeline

        # controller = BatchController(auto_process=False, kiara=self._kiara)

        pipeline = Pipeline(structure=self.structure)
        pipeline.inputs.set_values(**inputs)

        if not pipeline.inputs.items_are_valid():
            raise KiaraProcessingException(f"Can't start processing of {self._module_type_id} pipeline: one or several inputs missing or invalid.")  # type: ignore

        if not pipeline.status == StepStatus.RESULTS_READY:
            # TODO: error details
            raise KiaraProcessingException(f"Error when running pipeline of type '{self._module_type_id}'.")  # type: ignore

        outputs.set_values(**pipeline.outputs)



# -*- coding: utf-8 -*-


def test_example_controller():

    from kiara.examples.example_controller import (
        execute_pipeline_with_example_controller,
    )

    execute_pipeline_with_example_controller()



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel, Field

from kiara import Kiara
from kiara.data import Value
from kiara.data.onboarding import ValueStoreConfig
from kiara.operations import Operation


class BatchOnboard(BaseModel):
    @classmethod
    def create(
        cls,
        module_type: str,
        inputs: typing.Mapping[str, typing.Any],
        store_config: typing.Optional[ValueStoreConfig] = None,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        kiara: typing.Optional[Kiara] = None,
    ):

        if kiara is None:
            kiara = Kiara.instance()

        operation = Operation.create_operation(
            kiara=kiara,
            operation_id="_batch_onboarding_",
            config=module_type,
            module_config=module_config,
        )
        if store_config is None:
            store_config = ValueStoreConfig()
        elif isinstance(store_config, typing.Mapping):
            store_config = ValueStoreConfig(**store_config)

        return BatchOnboard(
            operation=operation, inputs=inputs, store_config=store_config
        )

    inputs: typing.Mapping[str, typing.Any] = Field(description="The inputs.")
    operation: Operation = Field(description="The operation to use with the inputs.")
    store_config: ValueStoreConfig = Field(
        description="The configuration for storing operation/pipeline values."
    )

    def run(self, base_id: str) -> typing.Mapping[Value, typing.Iterable[str]]:

        run_result = self.operation.module.run(**self.inputs)

        result: typing.Dict[Value, typing.List[str]] = {}
        if self.store_config.inputs:
            raise NotImplementedError("Storing inputs not supported yet.")
        if self.store_config.steps:
            raise NotImplementedError("Storing step inputs/outputs not supported yet.")
        if self.store_config.outputs:
            r = ValueStoreConfig.save_fields(
                base_id=base_id,
                value_set=run_result,
                matchers=self.store_config.outputs,
            )
            for value, aliases in r.items():
                result.setdefault(value, []).extend(aliases)

        return result



# -*- coding: utf-8 -*-
import typing
from pydantic import Field
from rich import box
from rich.console import RenderableType
from rich.panel import Panel
from rich.table import Table

from kiara import Kiara
from kiara.data.types import ValueTypeConfigMetadata
from kiara.info import KiaraInfoModel
from kiara.metadata import MetadataModel
from kiara.metadata.core_models import (
    ContextMetadataModel,
    DocumentationMetadataModel,
    MetadataModelMetadata,
    OriginMetadataModel,
    PythonClassMetadata,
)
from kiara.utils.output import create_table_from_base_model

if typing.TYPE_CHECKING:
    from kiara.data.types import ValueType


class ValueTypeInfo(KiaraInfoModel):
    @classmethod
    def from_type_class(
        cls, type_cls: typing.Type["ValueType"], kiara: typing.Optional[Kiara] = None
    ):

        type_attrs = cls.extract_type_attributes(type_cls=type_cls, kiara=kiara)
        return cls(**type_attrs)

    @classmethod
    def extract_type_attributes(
        self, type_cls: typing.Type["ValueType"], kiara: typing.Optional[Kiara] = None
    ) -> typing.Dict[str, typing.Any]:

        if kiara is None:
            kiara = Kiara.instance()

        origin_md = OriginMetadataModel.from_class(type_cls)
        doc = DocumentationMetadataModel.from_class_doc(type_cls)
        python_class = PythonClassMetadata.from_class(type_cls)
        properties_md = ContextMetadataModel.from_class(type_cls)

        value_type: str = type_cls._value_type_name  # type: ignore
        config = ValueTypeConfigMetadata.from_config_class(type_cls._config_class)

        metadata_keys = kiara.metadata_mgmt.get_metadata_keys_for_type(
            value_type=value_type
        )
        metadata_schemas: typing.Dict[str, typing.Type[MetadataModel]] = {}
        for metadata_key in metadata_keys:
            schema = kiara.metadata_mgmt.all_schemas.get(metadata_key)
            if schema is not None:
                metadata_schemas[metadata_key] = MetadataModelMetadata.from_model_class(
                    schema
                )

        return {
            "type_name": type_cls._value_type_name,  # type: ignore
            "documentation": doc,
            "origin": origin_md,
            "context": properties_md,
            "python_class": python_class,
            "config": config,
            "metadata_types": metadata_schemas,
        }

    type_name: str = Field(description="The name under which the type is registered.")
    documentation: DocumentationMetadataModel = Field(
        description="The documentation for this value type."
    )
    origin: OriginMetadataModel = Field(
        description="Information about authorship for the value type."
    )
    context: ContextMetadataModel = Field(
        description="Generic properties of this value type (description, tags, labels, references, ...)."
    )
    python_class: PythonClassMetadata = Field(
        description="Information about the Python class for this value type."
    )
    config: ValueTypeConfigMetadata = Field(
        description="Details on how this value type can be configured."
    )
    metadata_types: typing.Dict[str, MetadataModelMetadata] = Field(
        description="The available metadata types for this value type."
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        include_config_schema = config.get("include_config_schema", True)
        include_doc = config.get("include_doc", True)
        include_full_metadata = config.get("include_full_metadata", True)

        table = Table(box=box.SIMPLE, show_header=False, padding=(0, 0, 0, 0))
        table.add_column("property", style="i")
        table.add_column("value")

        if include_doc:
            table.add_row(
                "Documentation",
                Panel(self.documentation.create_renderable(), box=box.SIMPLE),
            )
        table.add_row("Origin", self.origin.create_renderable())
        table.add_row("Context", self.context.create_renderable())

        if include_config_schema:
            if config:
                config_cls = self.config.python_class.get_class()
                table.add_row("Type config", create_table_from_base_model(config_cls))
            else:
                table.add_row("Type config", "-- no config --")

        table.add_row("Python class", self.python_class.create_renderable())

        if include_full_metadata and self.metadata_types:
            md_table = Table(show_header=False, box=box.SIMPLE)
            for key, md in self.metadata_types.items():
                fields_table = md.create_fields_table(
                    show_header=False, show_required=False
                )
                md_table.add_row(f"[i]{key}[/i]", fields_table)

            table.add_row("Type metadata", md_table)

        return table



# -*- coding: utf-8 -*-
import collections
import typing

if typing.TYPE_CHECKING:
    from kiara.pipeline import StepValueAddress
    from kiara.pipeline.config import PipelineConfig
    from kiara.pipeline.pipeline import Pipeline
    from kiara.pipeline.structure import PipelineStructure


def extend_pipeline(
    pipeline: typing.Union["Pipeline", "PipelineStructure"],
    other: typing.Union[
        "Pipeline",
        "PipelineStructure",
        "PipelineConfig",
        typing.Mapping[str, typing.Any],
    ],
    input_links: typing.Optional[
        typing.Mapping[str, typing.Iterable["StepValueAddress"]]
    ] = None,
):

    from kiara.pipeline.config import PipelineConfig
    from kiara.pipeline.pipeline import Pipeline
    from kiara.pipeline.structure import PipelineStructure

    if isinstance(pipeline, Pipeline):
        structure: PipelineStructure = pipeline.structure
    elif isinstance(pipeline, PipelineStructure):
        structure = pipeline
    else:
        raise TypeError(f"Invalid type '{type(pipeline)}' for pipeline.")

    other_pipeline_config: typing.Optional[PipelineConfig] = None

    if isinstance(other, typing.Mapping):
        other_pipeline_config = PipelineConfig(**other)
    elif isinstance(other, PipelineConfig):
        other_pipeline_config = other
    elif isinstance(other, PipelineStructure):
        other_pipeline_config = other.structure_config

    if other_pipeline_config is None:
        from kiara.pipeline.pipeline import Pipeline

        if isinstance(other, Pipeline):
            other_pipeline_config = other.structure.structure_config

    if other_pipeline_config is None:
        raise TypeError(
            f"Can't extend pipeline structure, invalid type: {type(other)}."
        )

    other_structure: PipelineStructure = PipelineStructure(
        config=other_pipeline_config, kiara=structure._kiara
    )

    step_id_overlap = [
        el for el in set(structure.step_ids) if el in other_structure.step_ids
    ]
    if step_id_overlap:
        raise Exception(
            f"Can't extend pipeline, duplicate step id(s) are not allowed: {', '.join(step_id_overlap)}."
        )

    if input_links is None:
        input_links = {}
    else:
        input_links = dict(input_links)

    full_input_links: typing.Dict[
        str, typing.Dict[str, typing.Iterable[StepValueAddress]]
    ] = {}
    for input_name, input in other_structure.pipeline_inputs.items():
        if input_name not in input_links.keys():
            # this means we try to see whether there is an output in this structure with the same name
            if input_name in structure.pipeline_outputs.keys():
                output_step_address: typing.Iterable[StepValueAddress] = [
                    structure.pipeline_outputs[input_name].connected_output
                ]
            else:
                # output_step_address = "NEW INPUT"
                raise NotImplementedError()
        else:
            output_step_address = input_links[input_name]

        connected_inputs = other_structure.pipeline_inputs[input_name].connected_inputs
        if len(connected_inputs) != 1:
            raise NotImplementedError()
        else:
            connected_input = connected_inputs[0]
            full_input_links.setdefault(connected_input.step_id, {})[
                connected_input.value_name
            ] = output_step_address

    new_input_aliases = dict(structure._input_aliases)
    if not other_structure._output_aliases:
        new_output_aliases: typing.Union[str, typing.Mapping[str, str]] = "auto"
    else:
        new_output_aliases = dict(other_structure._output_aliases)

    config = structure.structure_config.dict(
        exclude={"input_aliases", "output_aliases", "steps"}
    )
    config["input_aliases"] = new_input_aliases
    config["output_aliases"] = new_output_aliases

    new_steps = [
        step.dict(exclude={"parent_id", "processing_stage", "required"})
        for step in structure.steps
    ]

    for step in other_structure.steps:
        step_dict = step.dict(exclude={"parent_id", "processing_stage", "required"})
        if step.step_id in full_input_links.keys():
            if step.step_id not in full_input_links.keys():
                new_steps.append(step_dict)
            else:
                for input_name in full_input_links[step.step_id].keys():
                    if input_name in step_dict["input_links"].keys():
                        raise NotImplementedError()
                    step_dict["input_links"][input_name] = full_input_links[
                        step.step_id
                    ][input_name]
                new_steps.append(step_dict)

    config["steps"] = new_steps
    pmc = PipelineConfig(**config)
    new_structure = pmc.create_pipeline_structure(kiara=structure._kiara)
    return new_structure


def generate_step_alias(step_id: str, value_name):
    return f"{step_id}.{value_name}"


def create_step_value_address(
    value_address_config: typing.Union[str, typing.Mapping[str, typing.Any]],
    default_field_name: str,
) -> "StepValueAddress":

    from kiara.pipeline.values import StepValueAddress

    if isinstance(value_address_config, StepValueAddress):
        return value_address_config

    sub_value: typing.Optional[typing.Mapping[str, typing.Any]] = None

    if isinstance(value_address_config, str):

        tokens = value_address_config.split(".")
        if len(tokens) == 1:
            step_id = value_address_config
            output_name = default_field_name
        elif len(tokens) == 2:
            step_id = tokens[0]
            output_name = tokens[1]
        elif len(tokens) == 3:
            step_id = tokens[0]
            output_name = tokens[1]
            sub_value = {"config": tokens[2]}
        else:
            raise NotImplementedError()

    elif isinstance(value_address_config, collections.abc.Mapping):
        print(value_address_config)
        step_id = value_address_config["step_id"]
        output_name = value_address_config["output_name"]
        sub_value = value_address_config.get("sub_value", None)
    else:
        raise TypeError(
            f"Invalid type for creating step value address: {type(value_address_config)}"
        )

    if sub_value is not None and not isinstance(sub_value, typing.Mapping):
        raise ValueError(
            f"Invalid type '{type(sub_value)}' for sub_value (step_id: {step_id}, value name: {output_name}): {sub_value}"
        )

    input_link = StepValueAddress(
        step_id=step_id, value_name=output_name, sub_value=sub_value
    )
    return input_link


def ensure_step_value_addresses(
    link: typing.Union[str, typing.Mapping, typing.Iterable], default_field_name: str
) -> typing.List["StepValueAddress"]:

    if isinstance(link, (str, typing.Mapping)):
        input_links: typing.List[StepValueAddress] = [
            create_step_value_address(
                value_address_config=link, default_field_name=default_field_name
            )
        ]

    elif isinstance(link, typing.Iterable):
        input_links = []
        for o in link:
            il = create_step_value_address(
                value_address_config=o, default_field_name=default_field_name
            )
            input_links.append(il)
    else:
        raise TypeError(f"Can't parse input map, invalid type for output: {link}")

    return input_links



# -*- coding: utf-8 -*-
import abc
import typing
from pydantic import Field

from kiara import KiaraModule
from kiara.data import ValueSet
from kiara.data.values import Value, ValueSchema
from kiara.exceptions import KiaraProcessingException
from kiara.module_config import ModuleTypeConfigSchema


class TypeConversionModuleConfig(ModuleTypeConfigSchema):

    source_type: str = Field(description="The source type.")
    target_type: str = Field(description="The target type.")


class OldTypeConversionModule(KiaraModule):

    _config_cls = TypeConversionModuleConfig

    @classmethod
    @abc.abstractmethod
    def _get_supported_source_types(self) -> typing.Union[typing.Iterable[str], str]:
        pass

    @classmethod
    @abc.abstractmethod
    def _get_target_types(self) -> typing.Union[typing.Iterable[str], str]:
        pass

    @classmethod
    def get_supported_source_types(self) -> typing.Set[str]:

        _types: typing.Iterable[str] = self._get_supported_source_types()
        if isinstance(_types, str):
            _types = [_types]

        if "config" in _types:
            raise Exception("Invalid source type, type name 'config' is invalid.")
        return set(_types)

    @classmethod
    def get_supported_target_types(self) -> typing.Set[str]:

        _types: typing.Iterable[str] = self._get_target_types()
        if isinstance(_types, str):
            _types = [_types]
        return set(_types)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    @property
    def source_type(self) -> str:
        data_type = self.get_config_value("source_type")
        supported = self.get_supported_source_types()
        if "*" not in supported and data_type not in supported:
            raise ValueError(
                f"Invalid module configuration, source type '{data_type}' not supported. Supported types: {', '.join(self.get_supported_source_types())}."
            )

        return data_type

    @property
    def target_type(self) -> str:
        data_type = self.get_config_value("target_type")
        if data_type not in self.get_supported_target_types():
            raise ValueError(
                f"Invalid module configuration, target type '{data_type}' not supported. Supported types: {', '.join(self.get_supported_target_types())}."
            )

        return data_type

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        inputs: typing.Mapping[str, typing.Any] = {
            "source_value": {
                "type": self.source_type,
                "doc": f"A value of type '{self.source_type}'.",
            },
            "config": {
                "type": "dict",
                "doc": "The configuration for the transformation.",
                "optional": True,
            },
        }

        return inputs

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        outputs = {
            "target_value": {
                "type": self.target_type,
                "doc": f"A value of type '{self.target_type}'.",
            }
        }
        return outputs

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value = inputs.get_value_obj("source_value")
        if value.value_schema.type != self.source_type:
            raise KiaraProcessingException(
                f"Can't convert value of source type '{value.value_schema.type}'. Expected type '{self.source_type}'."
            )
        config = inputs.get_value_data("config")
        if config is None:
            config = {}

        target_value = self.convert(value=value, config=config)
        # TODO: validate value?
        outputs.set_value("target_value", target_value)

    @abc.abstractmethod
    def convert(
        self, value: Value, config: typing.Mapping[str, typing.Any]
    ) -> typing.Any:
        pass



# -*- coding: utf-8 -*-
import abc
import typing
from pydantic import Field

from kiara import Kiara, KiaraModule
from kiara.data import ValueSet
from kiara.data.values import Value, ValueSchema
from kiara.exceptions import KiaraProcessingException
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message


class SampleValueModuleConfig(ModuleTypeConfigSchema):

    sample_type: str = Field(description="The sample method.")


class SampleValueModule(KiaraModule):

    _config_cls = SampleValueModuleConfig

    @classmethod
    @abc.abstractmethod
    def get_value_type(cls) -> str:
        """Return the value type for this sample module."""

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: Kiara
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        value_type: str = cls.get_value_type()

        if value_type not in kiara.type_mgmt.value_type_names:
            log_message(
                f"Ignoring sample operation for source type '{value_type}': type not available"
            )
            return {}

        for sample_type in cls.get_supported_sample_types():

            op_config = {
                "module_type": cls._module_type_id,  # type: ignore
                "module_config": {"sample_type": sample_type},
                "doc": f"Sample value of type '{value_type}' using method: {sample_type}.",
            }
            key = f"{value_type}.sample.{sample_type}"
            if key in all_metadata_profiles.keys():
                raise Exception(f"Duplicate profile key: {key}")
            all_metadata_profiles[key] = op_config

        return all_metadata_profiles

    @classmethod
    def get_supported_sample_types(cls) -> typing.Iterable[str]:

        types = []
        for attr_name, attr in cls.__dict__.items():

            if attr_name.startswith("sample_") and callable(attr):
                sample_type = attr_name[7:]
                if sample_type in types:
                    raise Exception(
                        f"Error in sample module '{cls.__name__}': multiple sample methods for type '{sample_type}'."
                    )
                types.append(sample_type)

        return types

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        return {
            "value_item": {
                "type": self.get_value_type(),
                "doc": "The value to sample.",
            },
            "sample_size": {
                "type": "integer",
                "doc": "The sample size.",
                "default": 10,
            },
        }

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        return {
            "sampled_value": {"type": self.get_value_type(), "doc": "The sampled value"}
        }

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        sample_size: int = inputs.get_value_data("sample_size")
        sample_type: str = self.get_config_value("sample_type")

        if sample_size < 0:
            raise KiaraProcessingException(
                f"Invalid sample size '{sample_size}': can't be negative."
            )

        value: Value = inputs.get_value_obj("value_item")

        func = getattr(self, f"sample_{sample_type}")
        result = func(value=value, sample_size=sample_size)

        outputs.set_value("sampled_value", result)


class SampleValueOperationType(OperationType):
    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, SampleValueModule)

    def get_operations_for_value_type(
        self, value_type: str
    ) -> typing.Dict[str, Operation]:
        """Find all operations that serialize the specified type.

        The result dict uses the serialization type as key, and the operation itself as value.
        """

        result: typing.Dict[str, Operation] = {}
        for o_id, op in self.operation_configs.items():
            sample_op_module_cls: typing.Type[SampleValueModule] = op.module_cls  # type: ignore
            source_type = sample_op_module_cls.get_value_type()
            if source_type == value_type:
                target_type = op.module_config["sample_type"]
                if target_type in result.keys():
                    raise Exception(
                        f"Multiple operations to sample '{source_type}' using {target_type}"
                    )
                result[target_type] = op

        return result



# -*- coding: utf-8 -*-

"""A command-line interface for *Kiara*.
"""
import click
import typing

from kiara import Kiara
from kiara.config import KiaraConfig
from kiara.utils import is_develop

from .data.commands import data
from .dev.commands import dev_group
from .explain import explain
from .info.commands import info
from .metadata.commands import metadata
from .module.commands import module
from .operation.commands import operation
from .pipeline.commands import pipeline
from .run import run
from .type.commands import type_group

# try:
#     import uvloop
#
#     uvloop.install()
# except Exception:
#     pass

# click.anyio_backend = "asyncio"


@click.group()
@click.option(
    "--pipeline-folder",
    "-p",
    help="Folder(s) that contain extra pipelines.",
    multiple=True,
    required=False,
)
@click.pass_context
def cli(ctx, pipeline_folder: typing.Tuple[str]):
    """Main cli entry-point, contains all the sub-commands."""

    ctx.obj = {}
    kc = KiaraConfig()
    if pipeline_folder:
        kc.extra_pipeline_folders = list(pipeline_folder)

    ctx.obj["kiara"] = Kiara(config=kc)


cli.add_command(explain)
cli.add_command(run)
cli.add_command(data)
cli.add_command(operation)
cli.add_command(module)
cli.add_command(pipeline)
cli.add_command(info)
cli.add_command(metadata)
cli.add_command(type_group)
if is_develop():
    cli.add_command(dev_group)

try:
    from .service.commands import service

    cli.add_command(service)
except ModuleNotFoundError:
    pass

# if is_develop():
#
#     from .dev.commands import dev
#
#     cli.add_command(dev)

if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
import pytest

import typing

from kiara import Kiara
from kiara.data.registry import DataRegistry, InMemoryDataRegistry
from kiara.data.values import NO_ID_YET_MARKER, Value, ValueSchema, ValueSlot


class SimpleTestRegistry(DataRegistry):
    """Test implementation of read-only registry, which uses parts of the value data as id."""

    def __init__(self, kiara: Kiara):

        self._values: typing.Dict[str, Value] = {}
        self._data: typing.Dict[str, typing.Any] = {}
        self._value_slots: typing.Dict[str, ValueSlot] = {}
        super().__init__(kiara=kiara)

    def _register_value_and_data(self, value: Value, data: typing.Any) -> str:

        assert isinstance(data, str)
        _id, aliases = data.split(".", maxsplit=1)

        if _id in self._values.keys():
            raise Exception(f"Duplicate id: {_id}")

        self._values[_id] = value
        self._data[_id] = data
        if aliases:
            for alias in aliases.split("."):
                if alias not in self._value_slots:
                    value_slot = ValueSlot(
                        id=alias,
                        value_schema=value.value_schema,
                        kiara=self._kiara,
                        registry=self,
                    )
                    self._value_slots[alias] = value_slot
                else:
                    value_slot = self._value_slots[alias]
                value_slot.add_value(value)
        return _id

    def _register_remote_value(self, value: Value) -> typing.Optional[Value]:
        raise NotImplementedError()

    def _get_available_value_ids(self) -> typing.Iterable[str]:
        return self._values.keys()

    def _get_value_obj_for_id(self, value_id: str) -> Value:
        return self._values[value_id]

    def _get_value_data_for_id(self, value_item: str) -> typing.Any:
        return self._data[value_item]

    def _get_available_aliases(self) -> typing.Iterable[str]:
        return self._value_slots.keys()

    def _get_value_slot_for_alias(self, alias_name: str) -> ValueSlot:
        return self._value_slots.get(alias_name, None)

    def _register_alias(self, alias_name: str, value_schema: ValueSchema) -> ValueSlot:
        raise NotImplementedError()


def test_read_registry_subclass(kiara: Kiara):

    reg = SimpleTestRegistry(kiara=kiara)

    string_schema = ValueSchema(type="string")
    value = reg.register_data(value_data="xxx.", value_schema=string_schema)
    assert value.id == "xxx"

    assert set(reg.alias_names) == set()

    with pytest.raises(Exception) as exc_info:
        reg.register_data(value_data="xxx.", value_schema=string_schema)

    assert "Duplicate" in str(exc_info.value)

    value_2 = reg.register_data(
        value_data="yyy.alias_1.alias_2", value_schema=string_schema
    )
    assert value_2.id == "yyy"
    assert set(reg.alias_names) == set(["alias_1", "alias_2"])

    assert len(reg.find_aliases_for_value("yyy")) == 2
    assert reg.get_versions_for_alias("alias_1") == [1]

    value_3 = reg.register_data(
        value_data="zzz.alias_1.alias_2", value_schema=string_schema
    )
    assert value_3.id == "zzz"
    assert reg.get_versions_for_alias("alias_1") == [1, 2]

    latest_alias_1_val = reg.get_value_obj("alias_1")
    assert latest_alias_1_val.id == "zzz"
    assert latest_alias_1_val.get_value_data() == "zzz.alias_1.alias_2"

    value_2_stored = reg.get_value_obj("alias_1@1")
    assert value_2 == value_2_stored
    assert value_2.get_value_data() == value_2_stored.get_value_data()


def test_default_data_registry(kiara: Kiara):

    reg = InMemoryDataRegistry(kiara=kiara)

    string_schema = ValueSchema(type="string")
    value_1 = reg.register_data("xxx", value_schema=string_schema)
    assert value_1.id
    assert value_1.id != NO_ID_YET_MARKER

    val_ret = reg.get_value_obj(value_1.id)
    assert val_ret == value_1
    assert val_ret.get_value_data() == value_1.get_value_data()

    assert not reg.alias_names
    assert value_1.id in reg.value_ids
    assert len(reg.value_ids) == 1

    value_2 = reg.register_data("xxx", value_schema=string_schema)
    assert value_2.get_value_data() == value_1.get_value_data()
    assert value_1.id != value_2.id

    assert len(reg.value_ids) == 2


def test_default_data_registry_aliases(kiara: Kiara):

    reg = InMemoryDataRegistry(kiara=kiara)

    string_schema = ValueSchema(type="string")
    value_slot_1 = reg.register_alias(value_or_schema=string_schema, alias_name="xxx")
    assert len(reg.alias_names) == 1
    assert "xxx" in reg.alias_names

    with pytest.raises(Exception) as exc_info:
        value_slot_1.get_latest_value()

    assert "No value added" in str(exc_info.value)

    value_ret = reg.get_value_obj("xxx")
    assert value_ret is None

    val_1 = reg.register_data(value_data="xxx", value_schema=string_schema)
    value_slot_1.add_value(val_1)

    assert value_slot_1.get_latest_value().get_value_data() == "xxx"

    assert val_1.id == value_slot_1.get_latest_value().id

    reg.register_alias(value_or_schema=string_schema, alias_name="yyy")
    aliases = reg.find_aliases_for_value(val_1.id)
    assert len(aliases) == 1
    assert aliases[0].alias == "xxx"



# -*- coding: utf-8 -*-
import pytest

import typing
from kiara_modules.core.logic import AndModule

from kiara import Kiara, KiaraModule


def test_module_instance_creation(kiara: Kiara):

    l_and = kiara.create_module("logic.and")
    assert l_and.config.dict() == {"constants": {}, "defaults": {}, "delay": 0}

    l_xor = kiara.create_module("logic.xor")
    assert "input_aliases" in l_xor.config.dict().keys()
    assert "output_aliases" in l_xor.config.dict().keys()

    with pytest.raises(Exception) as e_info:
        kiara.create_module("logic.and", module_config={"xxx": "fff"})

    assert "xxx" in str(e_info.value)
    assert "extra fields" in str(e_info.value)

    with pytest.raises(Exception) as e_info:
        kiara.create_module("logic.xor", module_config={"xxx": "fff"})

    assert "Can't dynamically create PipelineModuleClass" in str(e_info.value)


def test_module_instance_run(kiara: Kiara):

    l_and = kiara.create_module("logic.and")
    result = l_and.run(a=True, b=True)
    assert result.get_all_value_data() == {"y": True}

    with pytest.raises(Exception) as e_info:
        l_and.run()

    assert "input field(s) not valid" in str(e_info.value)

    with pytest.raises(Exception) as e_info:
        l_and.run(x=True)

    assert "Invalid input name" in str(e_info.value)

    l_xor = kiara.create_module("logic.xor")
    result = l_and.run(a=True, b=True)
    assert result.get_all_value_data() == {"y": True}

    with pytest.raises(Exception) as e_info:
        result = l_xor.run()
    assert "input field(s) not valid" in str(e_info.value)

    with pytest.raises(Exception) as e_info:
        l_xor.run(x=True)

    assert "Invalid input name" in str(e_info.value)


def test_module_instance_direct(kiara: Kiara):

    and_module = AndModule(kiara=kiara)
    result = and_module.run(a=True, b=True)
    assert result.get_all_value_data() == {"y": True}


def test_module_instance_from_class(kiara: Kiara):

    and_module = AndModule.create_instance(kiara=kiara)
    result = and_module.run(a=True, b=True)
    assert result.get_all_value_data() == {"y": True}

    with pytest.raises(Exception) as e:
        AndModule.create_instance(module_type="and", kiara=kiara)

    assert "but not both" in str(e.value)


def test_module_instance_from_config_file(
    kiara: Kiara, module_config_paths: typing.Dict[str, str]
):

    for path in module_config_paths.values():
        mod = KiaraModule.create_instance(path, kiara=kiara)
        assert isinstance(mod, KiaraModule)
        assert mod.input_schemas.keys()
        assert mod.output_schemas.keys()


def test_module_instance_from_pipeline_config_files(
    kiara: Kiara, pipeline_paths: typing.Dict[str, str]
):

    for path in pipeline_paths.values():
        mod = KiaraModule.create_instance(path, kiara=kiara)
        assert isinstance(mod, KiaraModule)
        assert mod.input_schemas.keys()
        assert mod.output_schemas.keys()
